/******************************************************************************************************************************
*          This program calculates Extreme Pathways using the algorithm given in
*          Schilling, C.H., Letscher, D, and Palsson, B.O. 2000. Theory for the Systemic
*          Definition of Metabolic Pathways and Their Use in Interpreting Metabolic Function
*          from a Pathway-oriented Perspective. J. Theoret. Biol. 203:229-248.
*
*          File: expa.c
*
*          Functions: main(), computePathways(), readMatrix(), readMatrix2(), numberOfNz(), makeBitArray(),
*                     printHelp(), reduceRows(), findColumn().
*
*          Descriptions: The function takes either a reaction file (of type .expa or .cel) as input or a
*          		 stoichiometric matrix (in the form of a text file) and outputs the extreme pathways in
*          		 a file named 'Paths.txt'.
*          		 Allowed calls: expa.exe sfile.txt xfile.txt nrows ncols nxflxs
*          		 where sfile.txt is the file containing the stoichiometric matrix
*          		       xfile.txt is the file containing the exchange fluxes in the following way:
*          		       each exchange flux is characterized as input, free, or output (-1,0,1). So, if there are
*          		       N exchange fluxes in the stoichiometric matrix, then xfile.txt contains N numbers, each number
*          		       either a -1, 0, or a 1. It is assumed that the last N columns of the stoichiometric matrix correspond
*          		       to the exchange fluxes.
*          		       nrows is the number of rows of the stoichiometric matrix
*          		       ncols is the number of columns of the stoichiometric matrix
*          		       nxflxs is the number of exchange fluxes (N in the example above).
*                      if there are no exchange fluxes (i.e., nxflxs = 0) then the argument 'xfile' is
*                      a name for a file which need not exist (the argument needs to be present, though).
*          		or expa.exe file.expa 
*          		where file.expa is a reaction file.
*          	Flags: if the flag `-s' is set, the program takes two arguments, including the flag
*                      the other argument is the name of the reaction file. 
*		       The program prints out the stoichiometric matrix to the terminal and then exits.
*		       if the flag `-v' is set, the program, in each iteration, prints out an upper bound for
*		       the number of rows in the next matrix (that upper bound is the quantity (zero + pos * neg)
*		       where zero is number of zero entries in the current column, pos is the number of positive
*		       entries in the current column, and neg is the number of negative entries. More precisely,
*		       each line in the output is a triple where the first number is a count, the second number
*		       is the column number (or row number in the stoichiometric matrix), and the third number is the
*		       upper bound.
*		       if the flag `-h' is set (with no other argument), the program prints out a help message and exits.
*		       If the flag -'version' is set, the program prints out the version number and exits
*		       If the flag '-c' is set the program will stop executing at the column given on the command
*		       line and print out the current path matrix.
*		       If the flag '-o' is set, the program will process the columns from a file in the order dictated
*		       by that file.
*		       If the flag '-p' is set, the program will process n columns and exit.
*		       If the flag '-r' is set, the program will process the column from a file.
*					   
*
*          Author: Steve Bell
*
*          Date: April 24, 2004.
*
************************************************************************************************************************
*
*          Revision 1.1 Dec 14, 2004. 
*          		moved full row manufacturing and insertion out of the inner (neg) loop
*          		eliminated all context switches (except for malloc and free) in the inner loop
*          		added reduceRows() and findColumn()
*          		added change of posPtrRow and negPtrRow
*
************************************************************************************************************************
*
*          Revision 1.2 Jan 24, 2005
*          		Sorted rows at beginning of each iteration
*          		Added much functionality to the program, including the
*          		ability to stop and start at a particular column
*
************************************************************************************************************************/



#define _EXPA_C

#include <time.h>
#include <stdio.h>
#include <stdlib.h>
#include <malloc.h>
#include <math.h> /*for log*/
#include <string.h>
#include "expa.h"
#include "queue.h" /*circular queue for column picking scheme*/
#include "util.h"
#include "readRxnFile.h"


static UINT32 T1nrows; /*number of rows in the current tableau*/
static UINT32 T2nrows; /*number of rows in the next tableau*/
static UINT32 TEnrows; /*number of rows in the matrix consisting of rows of the initial tableau corresponding to
			 *unconstrained exchange fluxes (TE)*/
static BOOLEAN stopDummy = my_false; /*set to true if stop at a particular column*/
static BOOLEAN colDummy = my_false; /*set to true if use ordered user defined columns */
static BOOLEAN partialDummy = my_false; /*set to true if compute only n columns */
static BOOLEAN useNewT1Ptr = my_false; /*set to true if input matrix and user defined columns is used*/
static BOOLEAN xDummy =my_false;
BOOLEAN stoickDummy = my_false;
BOOLEAN verbose = my_false; 

static time_t startTime;
static time_t endTime;
static UINT32 stopColumn = 0; /*column to stop at*/
static UINT32 nCols2Process = 0;

int main( int argc, char *argv[] )
{
	BOOLEAN done = my_false;
	BOOLEAN useStoickMat = my_false;
	UINT32* newCols = NULL; /*pointer to new columns if input matrix is used*/
	UINT32 switchVar, f, nFlags;
	UINT32 row = 0, col = 0, numberOfPathways, numberOfNewCols = 0;
	XTYPE ***stoickMatrix;
	XTYPE **T1Ptr, **T2Ptr, **TEPtr; /*pointer to current tableau, next tableau, and TE*/
	XTYPE ***T1PtrPtr, ***T2PtrPtr; /*keep track of pointers*/
	BIT_TYPE ***T1BitPtrPtr; /*keep track of pointer to bit array*/
	BIT_TYPE **T1BitPtr; /*pointer to bitArrays corresponding to T1*/
	
	XFLXTYPE* xFlxPtr; /* pointer to xflux type (INPUT = -1,FREE = 0,OUTPUT = 1) buffer */
	XFLXTYPE** xFlxPtrPtr; /* pointer to pointer to xflux type (INPUT,FREE,OUTPUT) buffer */
	char* filename = NULL;
	char* xflxFile = NULL;
	char* matFile = NULL;
	char* colFile = NULL;		
	char* flags = NULL;
	char flag;

	if( argc == 1 ){
		printf("Usage: expa [-choprvx] file1 [file2] [file3] [file4] [file5] [nrows] [ncols] [nxfluxes] [n] \n");
		printf("type 'expa -h' for help\n");
		exit(0);
	}

	flags = argv[1];
	if( flags[0] == '-' ){
		if( strcmp(flags, "-version") == 0 ){
			printf("expa version 1.1\n");
			exit(0);
		}

		nFlags = strlen(flags) - 1;
		for( f = 1; f <= nFlags; f++ ){
			flag = flags[f];
			switch(flag){
				case 'c':
					stopDummy = my_true;
					break;

				case 'h':
					printHelp();
					exit(0);

				case 'o':
					colDummy = my_true;
					break;

				case 'p':
					partialDummy = my_true;
					break;

				case 'r':
					useNewT1Ptr = my_true;
					break;

				case 's':
					stoickDummy = my_true;
					break;

				case 'v':
					verbose = my_true;
					break;

				case 'x':
					xDummy = my_true;
					break;

				default:
					fprintf(stderr, "Error in input argument %s \n", argv[1]);
					exit(3);
					break;

			}/*switch(flag)*/
		}/*for(f)*/
	}/*if( flags[0] == '-')*/

	if( (partialDummy == my_true) && (stopDummy == my_true) ){
		fprintf(stderr, "The flags -c and -p are not allowed together\n");
		exit(3);
	}
	if( (colDummy == my_true) && (useNewT1Ptr == my_true) ){
		fprintf(stderr, "The flags -o and -r are not allowed together\n");
		exit(3);
	}
	switchVar = argc;

	while ( done == my_false ){
		switch (switchVar) {
			case 2:/*no flags*/
				filename = argv[1];
				switchVar = 13;
				break;

			case 3:/*case: 'v'*/
				if( flags[0] == '-' ){
					if( (flags[1] != 'v') && (flags[1] != 's') && (flags[1] != 'x') ){
						fprintf(stderr, "Error in input argument %s\n", argv[1]);
						fprintf(stderr, "Usage: expa -[vsx] file.expa\n");
						exit(3);
					}
					filename = argv[2];
				}
				else{
					fprintf(stderr, "Error in input argument \n");
					fprintf(stderr, "Usage: expa -[vsx] file.expa\n");
					exit(3);
				}

				switchVar = 13;
				break;

			case 4:
				filename = argv[2];
				if( partialDummy == my_true ){/*case: 'p'*/
					if( stoickDummy == my_true ){
						fprintf(stderr, "The flags -p and -s are not allowed together\n");
						exit(3);
					}
					if( (nCols2Process = atoi(argv[3])) <= 0 ){;
						fprintf(stderr, "Error in input argument %s \n", argv[3]);
						fprintf(stderr, "Usage: expa -[v]p file.expa nCols\n");
						exit(3);
					}
				}
				else if( colDummy == my_true ){/*case: 'o' -- expa -[v]o file.expa cols.txt*/
					if( stoickDummy == my_true ){
						fprintf(stderr, "The flags -o and -s are not allowed together\n");
						exit(3);
					}
					filename = argv[2];
					colFile = argv[3];
				}
				else if( stopDummy == my_true ){/*case: 'c'*/
					if( stoickDummy == my_true ){
						fprintf(stderr, "The flags -c and -s are not allowed together\n");
						exit(3);
					}
					if( (stopColumn = atoi(argv[3])) <= 0 ){;
						fprintf(stderr, "Error in input argument %s \n", argv[3]);
						fprintf(stderr, "Usage: expa -[v]c file.expa stopCol\n");
						exit(3);
					}
				}
				else{
					fprintf(stderr, "Error in input argument \n");
					fprintf(stderr, "Usage: expa -[v][o][p][c] file.expa [cols.txt] [n]\n");
				}
				switchVar = 13;
				break;

			case 5:
				if( useNewT1Ptr == my_true ){/*case: 'r'*/
					if( stoickDummy == my_true ){
						fprintf(stderr, "The flags -r and -s are not allowed together\n");
						exit(3);
					}
					filename = argv[2];
					matFile = argv[3];
					colFile = argv[4];
				}
				else if( (colDummy == my_true) && (stopDummy == my_true ) ){/*case: 'o' and case: 'c'*/
					if( stoickDummy == my_true ){
						fprintf(stderr, "The flags -o and -s are not allowed together\n");
						exit(3);
					}
					filename = argv[2];
					colFile = argv[3];
					if( (stopColumn = atoi(argv[4])) <= 0 ){;
						fprintf(stderr, "Error in input argument %s \n", argv[4]);
						fprintf(stderr, "Usage: expa -[v]c file.expa cols.txt stopCol\n");
						exit(3);
					}
				}
				else if( (colDummy == my_true) && (partialDummy == my_true ) ){/*case: 'o' and case: 'p'*/
					if( stoickDummy == my_true ){
						fprintf(stderr, "The flags -o and -s are not allowed together\n");
						exit(3);
					}
					filename = argv[2];
					colFile = argv[3];
					if( (nCols2Process = atoi(argv[4])) <= 0 ){;
						fprintf(stderr, "Error in input argument %s \n", argv[4]);
						fprintf(stderr, "Usage: expa -[v]c file.expa cols.txt stopCol\n");
						exit(3);
					}
				}
				else{
					fprintf(stderr, "Usage: expa -[v][op][r] file.expa [mat.txt] [cols.txt] [nCol]\n");
					exit(3);
				}
				switchVar = 13;
				break;

			case 6:

				if( flags[0] == '-' ){
					if( useNewT1Ptr != my_true ){
						fprintf(stderr, "Error in argument: %s\n", flags);
						fprintf(stderr, "Usage: expar -[v][r][c][p] file.expa mat.txt cols.txt [n]\n");
						exit(3);
					}
					filename = argv[2];
					matFile = argv[3];
					colFile = argv[4];
					if( partialDummy == my_true ){/*case: 'p'*/
						if( stoickDummy == my_true ){/*case: 's'*/
							fprintf(stderr, "The flags -p and -s are not allowed together\n");
							exit(3);
						}
						if( (nCols2Process = atoi(argv[5])) <= 0 ){;
							fprintf(stderr, "Error in input argument %s \n", argv[5]);
							fprintf(stderr, "Usage: expar -[v][r][c][p] file.expa mat.txt cols.txt [n]\n");
							exit(3);
						}
					}
					else if( stopDummy == my_true ){/*case: 'c'*/
						if( stoickDummy == my_true ){
							fprintf(stderr, "The flags -c and -s are not allowed together\n");
							exit(3);
						}
						if( (stopColumn = atoi(argv[5])) <= 0 ){;
							fprintf(stderr, "Error in input argument %s \n", argv[5]);
							fprintf(stderr, "Usage: expar -[v][r][c][p] file.expa mat.txt cols.txt [n]\n");
							exit(3);
						}
					}
					switchVar = 13;
					break;
				}
				else{/*no flags*/
					if( flags[0] == '-' ){
						fprintf(stderr, "Error in argument, %s \n", argv[0]);
						printf("Usage: expa mat_file.txt xflx_file.txt nrows ncols nxflxs\n");
						exit(3);
					}
					filename = argv[1];
					xflxFile = argv[2];
					if( (NROWS = atoi(argv[3])) <= 0 ){
						fprintf(stderr, "Error in argument %s\n", argv[3]);
						printf("Usage: expa mat_file.txt xflx_file.txt nrows ncols nxflxs\n");
						exit(3);
					}
					if( (NCOLS = atoi(argv[4])) <= 0 ){
						fprintf(stderr, "Error in argument %s\n", argv[4]);
						printf("Usage: expa mat_file.txt xflx_file.txt nrows ncols nxflxs\n");
						exit(3);
					}
					if( (NXFLX = atoi(argv[5])) <= 0 ){
						fprintf(stderr, "Error in argument %s\n", argv[5]);
						printf("Usage: expa mat_file.txt xflx_file.txt nrows ncols nxflxs\n");
						exit(3);
					}
					xFlxPtr = xflxvector(NXFLX);
					useStoickMat = my_true;
					switchVar = 14;
					break;
				}

			case 7:
				if( flags[0] == '-' ){/*case: 'v'*/
					if( flags[1] != 'v'){
						fprintf(stderr, "Error in input argument %s\n", argv[1]);
						exit(3);
					}

					filename = argv[2];
					xflxFile = argv[3];
					if( (NROWS = atoi(argv[4])) <= 0 ){
						fprintf(stderr, "Error in argument %s\n", argv[4]);
						printf("Usage: expa -v mat_file.txt xflx_file.txt nrows ncols nxflxs\n");
						exit(3);
					}
					if( (NCOLS = atoi(argv[5])) <= 0 ){
						fprintf(stderr, "Error in argument %s\n", argv[5]);
						printf("Usage: expa -v mat_file.txt xflx_file.txt nrows ncols nxflxs\n");
						exit(3);
					}
					if( (NXFLX = atoi(argv[6])) <= 0 ){
						fprintf(stderr, "Error in argument %s\n", argv[6]);
						printf("Usage: expa -v mat_file.txt xflx_file.txt nrows ncols nxflxs\n");
						exit(3);
					}
					xFlxPtr = xflxvector(NXFLX);
				}
				else{
					fprintf(stderr, "Error in argument\n");
					printf("Usage: expa -v mat_file.txt xflx_file.txt nrows ncols nxflxs\n");
					exit(3);
				}
				useStoickMat = my_true;
				switchVar = 14;
				break;

			case 8:
				if( flags[0] == '-' ){
					filename = argv[2];
					xflxFile = argv[3];
					if( colDummy == my_false ){/*case: 'o'*/
						if( (NROWS = atoi(argv[4])) <= 0 ){
							fprintf(stderr, "Error in argument %s\n", argv[4]);
							fprintf(stderr, "Usage: expa -[v][o][p][c] mat_file.txt xflx_file.txt nrows ncols nxflxs stopCol\n");
							exit(3);
						}
						if( (NCOLS = atoi(argv[5])) <= 0 ){
							fprintf(stderr, "Error in argument %s\n", argv[5]);
							fprintf(stderr, "Usage: expa -[v][o][p][c] mat_file.txt xflx_file.txt nrows ncols nxflxs stopCol\n");
							exit(3);
						}
						if( (NXFLX = atoi(argv[6])) <= 0 ){
							fprintf(stderr, "Error in argument %s\n", argv[6]);
							fprintf(stderr,"Usage: expa -[v][o][p][c] mat_file.txt xflx_file.txt nrows ncols nxflxs stopCol\n");
							exit(3);
						}
						xFlxPtr = xflxvector(NXFLX);

						if( partialDummy == my_true ){/*case: 'p'*/
							if( stoickDummy == my_true ){
								fprintf(stderr, "The flags -p and -s are not allowed together\n");
								exit(3);
							}
							if( (nCols2Process = atoi(argv[7])) <= 0 ){;
								fprintf(stderr, "Error in input argument %s \n", argv[7]);
								fprintf(stderr,"Usage: expa -[v][o][p][c] mat_file.txt xflx_file.txt nrows ncols nxflxs ncols\n");
								exit(3);
							}
						}
						else if( stopDummy == my_true ){/*case 'c'*/
							if( stoickDummy == my_true ){
								fprintf(stderr, "The flags -c and -s are not allowed together\n");
								exit(3);
							}
							if( (stopColumn = atoi(argv[7])) <= 0 ){;
								fprintf(stderr, "Error in input argument %s \n", argv[7]);
								fprintf(stderr,"Usage: expa -[v][o][p][c] mat_file.txt xflx_file.txt nrows ncols nxflxs\n");
								exit(3);
							}
						}
						else{
							fprintf(stderr, "Error in input argument \n");
							fprintf(stderr,"Usage: expa -[v][o][p][c] mat_file.txt xflx_file.txt nrows ncols nxflxs stopCol\n");
						}
					}
					else{/*colDummy = my_true -- case: 'o'*/
						if( stoickDummy == my_true ){
							fprintf(stderr, "The flags -o and -s are not allowed together\n");
							exit(3);
						}
						colFile = argv[4];
						if( (NROWS = atoi(argv[5])) <= 0 ){
							fprintf(stderr, "Error in argument %s\n", argv[4]);
							fprintf(stderr, "Usage: expa -[v][o][p][c] mat_file.txt xflx_file.txt  nrows ncols nxflxs stopCol\n");
							exit(3);
						}
						if( (NCOLS = atoi(argv[6])) <= 0 ){
							fprintf(stderr, "Error in argument %s\n", argv[5]);
							fprintf(stderr, "Usage: expa -[v][o][p][c] mat_file.txt xflx_file.txt nrows ncols nxflxs stopCol\n");
							exit(3);
						}
						if( (NXFLX = atoi(argv[7])) <= 0 ){
							fprintf(stderr, "Error in argument %s\n", argv[6]);
							fprintf(stderr,"Usage: expa -[v][o][p][c] mat_file.txt xflx_file.txt nrows ncols nxflxs stopCol\n");
							exit(3);
						}
						xFlxPtr = xflxvector(NXFLX);
					}

				}
				else{/*if flags[0] == '-'*/
					fprintf(stderr, "Error in argument \n");
					fprintf(stderr,"Usage: expa -[v][o][p][c] mat_file.txt xflx_file.txt [col_file.txt] nrows ncols nxflxs [n]\n");
					exit(3);
				}
				useStoickMat = my_true;
				switchVar = 14;
				break;

			case 9:
				if( flags[0] == '-' ){
					if( (colDummy == my_true) && (stopDummy == my_true) ){/*case: 'o' and case: 'c'*/
						filename = argv[2];
						xflxFile = argv[3];
						if( stoickDummy == my_true ){
							fprintf(stderr, "The flags -o and -s or -c and -s are not allowed together\n");
							exit(3);
						}
						colFile = argv[4];
						if( (NROWS = atoi(argv[5])) <= 0 ){
							fprintf(stderr, "Error in argument %s\n", argv[5]);
							fprintf(stderr, "Usage: expa -[v]co mat_file.txt xflx_file.txt cols.txt nrows ncols nxflxs stopCol\n");
							exit(3);
						}
						if( (NCOLS = atoi(argv[6])) <= 0 ){
							fprintf(stderr, "Error in argument %s\n", argv[6]);
							fprintf(stderr, "Usage: expa -[v]co mat_file.txt xflx_file.txt cols.txt nrows ncols nxflxs stopCol\n");
							exit(3);
						}
						if( (NXFLX = atoi(argv[7])) <= 0 ){
							fprintf(stderr, "Error in argument %s\n", argv[7]);
							fprintf(stderr, "Usage: expa -[v]co mat_file.txt xflx_file.txt cols.txt nrows ncols nxflxs stopCol\n");
							exit(3);
						}
						if( (stopColumn = atoi(argv[8])) <= 0 ){;
							fprintf(stderr, "Error in input argument %s \n", argv[8]);
							fprintf(stderr,"Usage: expa -[v][o][p][c] mat_file.txt xflx_file.txt col_file.txt nrows ncols nxflxs stopCol\n");
							exit(3);
						}
						xFlxPtr = xflxvector(NXFLX);
					} /*if(colDummy && stopDummy)*/
					else if( (colDummy == my_true) && (partialDummy == my_true) ){/*case: 'o' and case: 'p'*/
						filename = argv[2];
						xflxFile = argv[3];
						if( stoickDummy == my_true ){
							fprintf(stderr, "The flags -o and -s or -c and -s are not allowed together\n");
							exit(3);
						}
						colFile = argv[4];
						if( (NROWS = atoi(argv[5])) <= 0 ){
							fprintf(stderr, "Error in argument %s\n", argv[5]);
							fprintf(stderr, "Usage: expa -[v]co mat_file.txt xflx_file.txt cols.txt nrows ncols nxflxs nCols\n");
							exit(3);
						}
						if( (NCOLS = atoi(argv[6])) <= 0 ){
							fprintf(stderr, "Error in argument %s\n", argv[6]);
							fprintf(stderr, "Usage: expa -[v]co mat_file.txt xflx_file.txt cols.txt nrows ncols nxflxs nCols\n");
							exit(3);
						}
						if( (NXFLX = atoi(argv[7])) <= 0 ){
							fprintf(stderr, "Error in argument %s\n", argv[7]);
							fprintf(stderr, "Usage: expa -[v]co mat_file.txt xflx_file.txt cols.txt nrows ncols nxflxs nCols\n");
							exit(3);
						}
						if( (nCols2Process = atoi(argv[8])) <= 0 ){;
							fprintf(stderr, "Error in input argument %s \n", argv[8]);
							fprintf(stderr,"Usage: expa -[v][o][p][c] mat_file.txt xflx_file.txt col_file.txt nrows ncols nxflxs nCols\n");
							exit(3);
						}
						xFlxPtr = xflxvector(NXFLX);
					} /*if(colDummy && stopDummy)*/
					else if( useNewT1Ptr == my_true ){/*case: 'r'*/
						filename = argv[2];
						xflxFile = argv[3];
						matFile = argv[4];
						colFile = argv[5];
						if( stoickDummy == my_true ){
							fprintf(stderr, "The flags -r and -s are not allowed together\n");
							exit(3);
						}
						if( (NROWS = atoi(argv[6])) <= 0 ){
							fprintf(stderr, "Error in argument %s\n", argv[6]);
							fprintf(stderr, "Usage: expa -[v]r stoick_file.txt xflx_file.txt mat.txt cols.txt nrows ncols nxflxs stopCol\n");
							exit(3);
						}
						if( (NCOLS = atoi(argv[7])) <= 0 ){
							fprintf(stderr, "Error in argument %s\n", argv[7]);
							fprintf(stderr, "Usage: expa -[v]r stoick_file.txt xflx_file.txt mat.txt cols.txt nrows ncols nxflxs stopCol\n");
							exit(3);
						}
						if( (NXFLX = atoi(argv[8])) <= 0 ){
							fprintf(stderr, "Error in argument %s\n", argv[8]);
							fprintf(stderr, "Usage: expa -[v]co stoick_file.txt xflx_file.txt mat.txt cols.txt nrows ncols nxflxs stopCol\n");
							exit(3);
						}
						xFlxPtr = xflxvector(NXFLX);
					}
				}
				else{/*if flags[0] == '-')*/
					fprintf(stderr, "Error in argument\n");
					fprintf(stderr, "Usage: expa - [v][r][c][o] stoick_file.txt xlfx_file.txt [mat.txt] cols.txt nrows ncols nxflxs [stopCol]\n");
					exit(3);
				}
				useStoickMat = my_true;
				switchVar = 14;
				break;

			case 10:
				if( useNewT1Ptr == my_true ){/*case: 'r'*/
					filename = argv[2];
					xflxFile = argv[3];
					matFile = argv[4];
					colFile = argv[5];
					if( stoickDummy == my_true ){
						fprintf(stderr, "The flags -r and -s are not allowed together\n");
						exit(3);
					}
					if( (NROWS = atoi(argv[6])) <= 0 ){
						fprintf(stderr, "Error in argument %s\n", argv[6]);
						fprintf(stderr, "Usage: expar -[v]r stoick_file.txt xflx_file.txt mat.txt cols.txt nrows ncols nxflxs stopCol\n");
						exit(3);
					}
					if( (NCOLS = atoi(argv[7])) <= 0 ){
						fprintf(stderr, "Error in argument %s\n", argv[7]);
						fprintf(stderr, "Usage: expar -[v]r stoick_file.txt xflx_file.txt mat.txt cols.txt nrows ncols nxflxs stopCol\n");
						exit(3);
					}
					if( (NXFLX = atoi(argv[8])) <= 0 ){
						fprintf(stderr, "Error in argument %s\n", argv[8]);
						fprintf(stderr, "Usage: expar -[v]co stoick_file.txt xflx_file.txt mat.txt cols.txt nrows ncols nxflxs stopCol\n");
						exit(3);
					}
					xFlxPtr = xflxvector(NXFLX);

					if( partialDummy == my_true ){/*case: 'p'*/
						if( stoickDummy == my_true ){
							fprintf(stderr, "The flags -p and -s are not allowed together\n");
							exit(3);
						}
						if( (nCols2Process = atoi(argv[9])) <= 0 ){;
							fprintf(stderr, "Error in input argument %s \n", argv[9]);
							fprintf(stderr,"Usage: expar -[v][p][c] mat_file.txt xflx_file.txt nrows ncols nxflxs ncols\n");
							exit(3);
						}
					}
					else if( stopDummy == my_true ){/*case: 'c'*/
						if( stoickDummy == my_true ){
							fprintf(stderr, "The flags -c and -s are not allowed together\n");
							exit(3);
						}
						if( (stopColumn = atoi(argv[9])) <= 0 ){;
							fprintf(stderr, "Error in input argument %s \n", argv[9]);
							fprintf(stderr,"Usage: expar -[v][p][c] mat_file.txt xflx_file.txt mat.txt cols.txt nrows ncols nxflxs stopCol\n");
							exit(3);
						}
					}
					else{
						fprintf(stderr, "Error in input argument \n");
						fprintf(stderr,"Usage: expar -[v][p][c] mat_file.txt xflx_file.txt nrows ncols nxflxs n\n");
					}
				}
				else{
					fprintf(stderr, "Error in input argument\n");
					fprintf(stderr,"Usage: expar -[v][p][c] stoick_file.txt xflx_file.txt mat.txt col_file.txt nrows ncols nxflxs stopCol\n");
				}
				useStoickMat = my_true;
				switchVar = 14;
				break;

			case 13:
				if( ( xFlxPtrPtr = (XFLXTYPE**) malloc((unsigned) sizeof(XFLXTYPE*)) ) == NULL ){
					fprintf(stderr, "Cannot allocate pointer for xFlxPtrPtr \n");
					exit(4);
				}
				if( ( stoickMatrix = (XTYPE***) malloc((unsigned) sizeof(XTYPE**)) ) == NULL ){
					fprintf(stderr, "Cannot allocate pointer for stoick matrix \n");
					exit(4);
				}
				
				readRxnFile(filename, stoickMatrix, xFlxPtrPtr, &NROWS, &NCOLS, &NXFLX);
				if( stoickDummy == my_true ){
					writeStoickMat(*stoickMatrix, NROWS, NCOLS);
					freextypeMatrix(*stoickMatrix, NROWS);
					free((void*) stoickMatrix);	
					if( stoickDummy == my_true ){
						exit(0);
					}
				}
				if( xDummy ==  my_true ){
					for( row = 1; row <= NXFLX; row++ ){
						printf("%d\n", (*xFlxPtrPtr)[row]);
					}
					exit(0);
				}

				switchVar = 14;
				break;

			case 14:

				T1PtrPtr = (XTYPE***) malloc( (unsigned) sizeof(XTYPE**) );
				T2PtrPtr = (XTYPE***) malloc( (unsigned) sizeof(XTYPE**) );
				if( ( T1PtrPtr == NULL ) || ( T2PtrPtr == NULL ) ){
					nrerror("Failed to allocate pointer");
				}

				T1BitPtrPtr = (BIT_TYPE***) malloc( (unsigned) sizeof(BIT_TYPE**) );
				if( T1BitPtrPtr == NULL ){
					nrerror("Failed to allocate pointer: T1BitPtrPtr");
				}
					
				if( QueueInit(NROWS) == my_false ){ /*initialize the circular queue*/
					fprintf(stderr, "Cannot initialize queue \n");
					exit(4);
				}

				
				T1Ptr = (XTYPE**) malloc( (unsigned) (NCOLS + 1) * sizeof( XTYPE* ) );
				T1nrows = NCOLS;
				
				T2Ptr = (XTYPE**) malloc( (unsigned) (NCOLS + 1) * sizeof( XTYPE* ) );
				if( T2Ptr == NULL ){
					nrerror("Failed to allocate pointer:T2Ptr");
				}
				*T2PtrPtr = T2Ptr;

				TEPtr = xtypeMatrix(NCOLS, NROWS + NCOLS);	
				TEnrows = NCOLS;

				NUMBER_OF_XTYPES = (NROWS + NCOLS) / XTYPESIZE;
				NUMBER_OF_RESIDS = (NROWS + NCOLS) % XTYPESIZE;

				T1BitPtr = bitTypeMatrix(NCOLS, NUMBER_OF_XTYPES + 1);
				if( T1BitPtr == NULL ){
					nrerror("Failed to allocate pointer: T1BitPtr");
				}
				if( useStoickMat == my_true ){
					switchVar = 15;
				}
				else{
					switchVar = 16;
				}
				break;

			case 15:
				readMatrix(T1Ptr, TEPtr, T1BitPtr, xFlxPtr, filename, xflxFile);
				if( useNewT1Ptr == my_true ){
					for( row = 1; row <= T1nrows; row++ ){
						freebitTypeVector((void*) T1BitPtr[row]);
						freeVector((void*) T1Ptr[row]);
					}
					free((void*) T1BitPtr);
					free((void*) T1Ptr);
					newCols = uint32vector(NROWS);
					makeColPtr(newCols, colFile, &numberOfNewCols);
					readMatrix3(T1PtrPtr, T1BitPtrPtr, matFile);
				}
				else if( colDummy == my_true ){
					newCols = uint32vector(NROWS);
					makeColPtr(newCols, colFile, &numberOfNewCols);
				}
				done = my_true;
				break;

			case 16:
				readMatrix2(T1Ptr, TEPtr, T1BitPtr, *xFlxPtrPtr, stoickMatrix);
				if( useNewT1Ptr == my_true ){
					for( row = 1; row <= T1nrows; row++ ){
						freebitTypeVector((void*) T1BitPtr[row]);
						freeVector((void*) T1Ptr[row]);
					}
					free((void*) T1BitPtr);
					free((void*) T1Ptr);
					newCols = uint32vector(NROWS);
					makeColPtr(newCols, colFile, &numberOfNewCols);
					readMatrix3(T1PtrPtr, T1BitPtrPtr, matFile);
				}
				else if( colDummy == my_true ){
					newCols = uint32vector(NROWS);
					makeColPtr(newCols, colFile, &numberOfNewCols);
				}
				freextypeMatrix(*stoickMatrix, NROWS);
				free((void*) stoickMatrix);	
				done = my_true;
				break;

			default:	
				fprintf(stderr,"Usage: expa [-choprv] file1 [file2] [file3] [file4] [file5] [nrows] [ncols] [nxfluxes] [n] \n");
				fprintf(stderr,"type 'expa -h' for help\n");
				exit(2);
		}
	}


	if( useNewT1Ptr == my_false ){	
		*T1PtrPtr = T1Ptr;
		*T1BitPtrPtr = T1BitPtr;
	}
	if( verbose == my_true ){
		time(&startTime);
	}	
	
	computePathways(T1PtrPtr, T2PtrPtr, TEPtr, T1BitPtrPtr, newCols, numberOfNewCols); 

	if( verbose == my_true ){
		time(&endTime);
	}

	reduceRows(*T1PtrPtr, *T1BitPtrPtr, T1nrows);	
	writeMat(*T1PtrPtr, *T1BitPtrPtr, T1nrows);


	if( (argc == 6) || (argc == 7) || (argc == 9) || (argc == 10) || ((argc == 8) && (colDummy == my_true)) ){	
		freexflxVector(xFlxPtr);
	}
	else{
		freexflxVector(*xFlxPtrPtr);
		free((void*) xFlxPtrPtr);
	}
	
	numberOfPathways = 0;
	for( row = 1; row <= T1nrows; row++ ){
		if( (*T1PtrPtr)[row] != NULL ){
			freeVector( (*T1PtrPtr)[row] );
			numberOfPathways++;
		}
		if( (*T1BitPtrPtr)[row] != NULL ){
			freebitTypeVector( (*T1BitPtrPtr)[row] );
		}
	}

	if( verbose == my_true ){
		printf("\n\nnumber of pathways: %d\n", numberOfPathways);
		printf("Time in seconds: %ld \n", endTime - startTime);
		printf("Exiting ... file is Paths.txt\n");
	}
	
	free((void*) *T1PtrPtr);
	free((void*) *T2PtrPtr);
	free((void*) *T1BitPtrPtr);
	freextypeMatrix(TEPtr, TEnrows);
	free((void*) T1PtrPtr);
	free((void*) T2PtrPtr);
	free((void*) T1BitPtrPtr);
	return(0);
} /*main()*/
/*****************************************************************************************************************************************
 *
 * 		Function: computePathways
 *
 * 		Description: This function computes the extreme pathways. The inputs are the initial
 * 			     tableau (minus the rows corresponding to the unconstrained exchange fluxes),
 * 			     the matrix containg the rows corresponding to the unconstrained exchange fluxes,
 * 			     and the bit map representation of the initial tableau. The bit map representation
 * 			     is used since only the non-zero elements of the initial tableau are stored in memory. 
 * 			     This representation then keeps track of which elements in a particular row of the initial tableau
 * 			     is non-zero by setting bits corresponding to non-zero elements to 1, and 0 otherwise starting
 * 			     with the most significant bit. For example, if the number of columns in the initial tableau is
 * 			     35, then the bit map will have two words in each row (assuming that a word is 32 bits), word1 and
 * 			     word2, say, where word1 is the leftmost word. Suppose now that that the first entry in a particular row
 * 			     of the initial tableau is non-zero. Then the most significant bit (bit number 32, counting from right to left)
 * 			     will be set. Likewise, if the last entry in a row is non-zero, then bit number 30 (or the third bit from the
 * 			     left) will be set in word2 of the bitmap. The remaining bits are set to zero. (A mask is used to keep track
 * 			     of the significant bits of the last word in the bit map.)
 * 			     The column to be zeroed out in each iteration is chosen as follows. For each row, the number of zero
 * 			     entries plus the product of the number of positive entries and the number of negative entries is recorded.
 * 			     The first column which achieves the minimum of this quantity is selected as the pivoting column. 
 * 			     
 *
 * 		Parameters: mat1PtrPtr -- pointer to the initial tableau, also used as a temporary matrix
 * 			    for forming the next tableau in the main loop.
 * 			    mat2PtrPtr -- pointer to the final tableau containining the extreme pathways, also
 * 			    used as a tableau in each iteration.
 * 			    matEPtr -- matrix containing the rows of the initial tableau which correspond
 * 			    to unconstrained exchange fluxes (referred to as T^E in Schilling et al.).
 * 			    bit1PtrPtr -- pointer to the bit map representation of the initial tableau (as well
 * 			    as each of the other tableaux). 
 *
 * 		Date:       April 24, 2004
 *
 * 		Author:	    Steve Bell
 *
 * **************************************************************************************************************************************/


void computePathways(XTYPE*** mat1PtrPtr, XTYPE*** mat2PtrPtr, XTYPE** matEPtr, BIT_TYPE*** bit1PtrPtr, UINT32* newColumns,
								UINT32 numberOfNewColumns )
{
	/*used in check independence*/
	UINT32 nnewAllocatedRows, numberOfTempRows, numberOfRows, nAllocatedTempRows, row2;
	UINT32* nextPosNegPtr;
	BIT_TYPE* nextRowBitPtr;	
	BIT_TYPE** tempBitPtr;
	BIT_TYPE** newRowsBitPtr;
	UINT32** newRowsPosNegPtr;
	UINT32** tempPosNegPtr;

	BOOLEAN memAllocDummy = my_false; /*when memory allocation gets too big, use this dummy variable to incrementally
				      *allocate memory*/
	/*used for sorting rows*/
	UINT32* sortedPtrs;
	UINT32 nz2bSorted, row3, row4, nSortedAllRows;

	/* used for user defined columns*/
	UINT32 colIndex, numberOfCols;

	/*used for partial execution*/
	UINT32 nColsProcessed = 0;
	BOOLEAN newColDummy = my_false;
	/*used for finding minimum column*/
	UINT32 dummy = 0, queueSize, minNR; 
	UINT32 *minColPtr;
	UINT32 row, col, xtype, columnNumber = 0; /*dummy variables*/
	UINT32 t2nrows; /*number of rows in second matrix*/
	UINT32 t1nallRows, t2nallRows, tempRows, posNegAllRows; /*allocated rows*/
	/*these are pointers to matrices and bit arrays*/
	XTYPE **tempPtr, **mat1Ptr, **mat2Ptr;
	BIT_TYPE **tBitPtr, **bit2Ptr, **bit1Ptr;	
	/*these are used to find a particular column in the bit array*/ 
	UINT32 colSlot;
	BIT_TYPE colBit;
	UINT32 nz, nzCol; /*these are used to find non zero rows */
	/*pointers for new row to be tested for conical independence*/
	BIT_TYPE *newRowBitPtr;
	XTYPE *newRowPtr;
    /*for a particular column, these keep track of which rows have zero, nonZero, positive, negative
	 * entries*/	
	UINT32 zero, neg, pos, npos, nneg, nonZero, cnewRows, nnewRows, nzRow, tempNeg; 
	UINT32 *posPtrRow, *posPtrCol, *negPtrRow, *negPtrCol, *nonZeroPtr, *tPosNegPtr;
	/*these are used in 'identify unconstrained metabolites'*/
	UINT32 uSize = 0, u = 1,  *uxflxPtr;
	BOOLEAN uflx = my_false; /*exchange flux*/
	UINT32 countNz, countCol; /*these are used to find columns in sparse vectors*/  
	BIT_TYPE xWord;
	/*these are used in 'combine rows'*/	
	BOOLEAN STOP = my_false;
	XTYPE negCoeff, posCoeff;
	UINT32 bitCount, posBit, negBit, negCol, posCol, negRow, posRow, nextBit, bits, bitStop;
	UINT32 posSlot, negSlot, newCol, vectorSize;
	/*these are used in 'add TE'*/
	UINT32 TEcol, TErow, mat1Slot, newSlot;
	BIT_TYPE testBit, tempBit;
	XTYPE matEValue, T1coeff;
	/*these are used in 'test row'*/
	BIT_TYPE mask;
	UINT32 numberOfXtypes = NCOLS/(sizeof(XTYPE) * 8),  bit;
	UINT32 numberOfResiduals = NCOLS%(sizeof(XTYPE) * 8);
	BOOLEAN addRow, delRow; 
	/*this are used in 'find row size'*/
	BIT_TYPE highBit, lowBit;
	/*main matrices used*/
	mat1Ptr = *mat1PtrPtr;
	mat2Ptr = *mat2PtrPtr;
	bit1Ptr = *bit1PtrPtr;
	/*pointer to minimum columns*/
	minColPtr = uint32vector(NROWS);
	/*pointer to mets which have unconstrained xfluxes*/
	uxflxPtr = uint32vector(NROWS + NCOLS);
	/*pointer to the rows which have positive entries in the column being investigated*/
	posPtrRow = uint32vector(NROWS + NCOLS);
	/*pointer to the rows which have negative entries in the column being investigated*/
	negPtrRow = uint32vector(NROWS + NCOLS);
	/*pointer to the columns in the sparse matrix whose rows have positive entries in the column being investigated*/
	posPtrCol = uint32vector(NROWS + NCOLS);
	/*pointer to the columns in the sparse matrix whose rows have negative entries in the column being investigated*/
	negPtrCol = uint32vector(NROWS + NCOLS);
	/*pointer to the columns in the sparse matrix whose rows have non zero entries in the column being investigated*/
	nonZeroPtr = uint32vector(NROWS + NCOLS);
	/*pointer to the equivalent sets -- */


	t1nallRows = T1nrows;
	t2nallRows = NCOLS;
	nSortedAllRows = NCOLS;
	posNegAllRows = NROWS + NCOLS;
	nAllocatedTempRows = NROWS + NCOLS;
	nnewAllocatedRows = NROWS + NCOLS;

	/*pointer to bitarray*/
	bit2Ptr = (BIT_TYPE**) malloc( (unsigned) (t2nallRows + 1) * sizeof( BIT_TYPE* ) );
	if( bit2Ptr == NULL ){
		nrerror("Failed to allocate pointer: bit2Ptr");
	}

	/*pointer to sorted rows*/
	if( (sortedPtrs = (UINT32*) malloc( (unsigned) nSortedAllRows * sizeof( UINT32 ) ) ) == NULL ){
		fprintf(stderr, "cannot allocate sortedPtrs\n");
		exit(1);
	}
	sortedPtrs--;

	/*newRowsBitPtr holds the bit reps and newRowsPosNegPtr holds the positive/negative rowindices in check independence */
	if( (newRowsBitPtr = (BIT_TYPE**) malloc( (unsigned) (nnewAllocatedRows + 1) * sizeof(BIT_TYPE*))) == NULL ){
		fprintf(stderr, "cannot reallocate newRowsBitPtr\n");
		exit(1);
	}

	if( (newRowsPosNegPtr = (UINT32**) malloc((unsigned) (nnewAllocatedRows + 1) * sizeof(UINT32*))) == NULL ){
		fprintf(stderr, "cannot allocate memory for newRowsPosNegPtr\n");
		exit(1);
	}
	numberOfRows = 0;


	/*temporary holder for bit rows*/
	if( (tempBitPtr = (BIT_TYPE**) malloc((unsigned) (nAllocatedTempRows + 1) * sizeof(BIT_TYPE*))) == NULL ){
		fprintf(stderr, "cannot allocate tempBitPtr\n");
		exit(1);
	}
	/*temporary holder for pos-neg rows*/
	if( (tempPosNegPtr = (UINT32**) malloc((unsigned) (nAllocatedTempRows + 1) * sizeof(UINT32*))) == NULL ){
		fprintf(stderr, "cannot allocate memory for tempPosNegPtr\n");
		exit(1);
	}


	/*used in test for conical independence*/
	mask = HIGH_BIT;
	tempBit = mask;
	delRow = my_false;
	for( bit = 1; bit <= numberOfResiduals; bit++ ){
		mask = mask | tempBit;
		tempBit = tempBit >> 1;
		
	}
    
	if( colDummy == my_true ){
		numberOfCols = numberOfNewColumns;
	}
	else{
		numberOfCols = 0;
	}

	/*identify unconstrained metabolites*********************************************************************************/
	u = 1;
	uSize = 0; /*used in add TE*/
	for( col = NCOLS + 1; col <= NROWS + NCOLS; col++ ){
		for( row = 1; row <= TEnrows; row++ ){
			if( ( matEPtr != NULL ) && ( matEPtr[row][col] != 0 ) ){
				uflx = my_true;
				break;
			}
		}
		if( uflx == my_true ){
			uxflxPtr[u] = col;
			u++;
			uSize++;
			uflx = my_false;
		}
		else{
		    /*mets that don't have an unconstrained xflux*/
			if( useNewT1Ptr == my_false ){
				newColDummy = my_false;
				for( colIndex = 1; colIndex <= numberOfCols; colIndex++ ){
					if( col == newColumns[colIndex] ){
						newColDummy = my_true;
						break;
					}
				}
				if( newColDummy == my_true ){
					continue;
				}
				if( PutWord( col ) == QUEUE_FULL ){
					fprintf(stderr, "error in input file -- number of columns = ?");
					exit(4);
				}
			}
		}
	}
	/*end identify unconstrained metabolites*********************************************************************************/

	/*if use user defined columns*/
	if( useNewT1Ptr == my_true ){
		for( col = 1; col <= numberOfNewColumns; col++ ){
			PutWord( newColumns[col] );
		}
	}

	if( (useNewT1Ptr == my_true) || (colDummy == my_true) ){
		for( row3 = 1; row3 <= numberOfNewColumns; row3++ ){
			for( row4 = 1; row4 <= uSize; row4++ ){
				if( newColumns[row3] == uxflxPtr[row4] ){
					fprintf(stderr, "Error: columns with FREE exchange fluxes are not allowed in column file: %ld\n",
							(newColumns[row3] - NCOLS));
					exit(3);
				}
			}
		}
	}

	colIndex = 1;
	while(1){/*main loop*/
	
		minNR = MAXBITS;
		dummy = 0; /*used to find minimum column*/
		queueSize = GetNumberOfWords();
		if( queueSize == 0 ){
			break;
		}
		
		/*find column which has the minimum number of positive rows * number of negative rows*/
		if( colIndex <= numberOfCols ){
			col = newColumns[colIndex];
			colIndex++;
		}
		else{
			col = findColumn(bit1Ptr, mat1Ptr, queueSize);
		}
	
		neg = 0;
		pos = 0;
		zero = 0;
		nonZero = 0;
	
		/*expand matrix if needed*/
		if( t2nallRows + 1 < T1nrows ){
			t2nallRows = T1nrows;
			mat2Ptr = (XTYPE**) realloc((void*) mat2Ptr, (unsigned) (t2nallRows + 1) * sizeof( XTYPE* ) );
			if( mat2Ptr == NULL ){
				nrerror("Failed to allocate pointer: mat2Ptr");
			}
			bit2Ptr = (UINT32**) realloc( (void*) bit2Ptr, (unsigned) (t2nallRows + 1) * sizeof( UINT32* ) );
			if( bit2Ptr == NULL ){
				nrerror("Failed to allocate pointer: bit2Ptr");
			}
		}
		if( posNegAllRows + 1 < T1nrows ){
			posNegAllRows += T1nrows;
			posPtrRow = (UINT32*) realloc( (void*) posPtrRow, (unsigned) ( posNegAllRows + 1) * sizeof( UINT32 ) );
			negPtrRow = (UINT32*) realloc( (void*) negPtrRow, (unsigned) ( posNegAllRows + 1) * sizeof( UINT32 ) );
			posPtrCol = (UINT32*) realloc( (void*) posPtrCol, (unsigned) ( posNegAllRows + 1) * sizeof( UINT32 ) );
			negPtrCol = (UINT32*) realloc( (void*) negPtrCol, (unsigned) ( posNegAllRows + 1) * sizeof( UINT32 ) );
			nonZeroPtr = (UINT32*) realloc( (void*) nonZeroPtr, (unsigned) ( posNegAllRows + 1) * sizeof( UINT32 ) );
			if( ( posPtrRow == NULL ) || ( negPtrRow == NULL ) || ( posPtrCol == NULL ) 
				|| (negPtrCol == NULL) || (nonZeroPtr == NULL)		){
					nrerror("Failed to allocate pointer");
			}
		}

		if( nSortedAllRows < T1nrows ){
			nSortedAllRows = T1nrows;
			sortedPtrs++;
			if( (sortedPtrs = (UINT32*) realloc( (void*) sortedPtrs, (unsigned) 
							nSortedAllRows * sizeof( UINT32 ) ) ) == NULL ){
				nrerror("Failed to reallocate pointer: sortedPtrs");
			}
			sortedPtrs--;
		}

		/*find current column in bit array*/
		colSlot = (col - 1)/BIT_TYPESIZE  + 1;
		if( (col - 1) % BIT_TYPESIZE  != 0 ){
	       		colBit =  HIGH_BIT >> ( (col - 1) % BIT_TYPESIZE );
		}
		else{
			colBit = HIGH_BIT;
		}

		/*copy rows with zeros in col to next tableau*/
		for( row = 1; row <= T1nrows; row++ ){
			if( bit1Ptr[row] == NULL ){
				continue;
			}
			if( (bit1Ptr[row][colSlot] & colBit) == 0 ){
				zero++;
				/*find size of row*/
				nz = 0;
				highBit = HIGH_BIT;
				lowBit = LOW_BIT;
				for( xtype = 1; xtype <= numberOfXtypes; xtype++ ){
					tempBit = bit1Ptr[row][xtype];
					for( bits = 0; bits < BIT_TYPESIZE; bits++ ){
						if( (tempBit & lowBit) != 0 ){
							nz++;
						}
						tempBit = tempBit>>1;
					}
				}
				tempBit = bit1Ptr[row][xtype];
				for( bits = 0; bits < numberOfResiduals; bits++){
					if( (tempBit & highBit) != 0 ){
						nz++;
					}
					tempBit = tempBit<<1;
				}
				/*end find size -- size = nz*/

				nz2bSorted = nz;
				for( row3 = 1; row3 <= (zero - 1); row3++ ){
					if( nz2bSorted >= sortedPtrs[row3] ){
						for( row4 = zero; row4 >= (row3 + 1); row4-- ){
							sortedPtrs[row4] = sortedPtrs[row4 - 1];
							mat2Ptr[row4] = mat2Ptr[row4 - 1];
							bit2Ptr[row4] = bit2Ptr[row4 - 1];
						}
						sortedPtrs[row4] = nz2bSorted;
						mat2Ptr[row4] = mat1Ptr[row];
						bit2Ptr[row4] = bit1Ptr[row];
						break;
					}
				}
				if( row3 == zero ){
					sortedPtrs[zero] = nz2bSorted;
					mat2Ptr[zero] = mat1Ptr[row];
					bit2Ptr[zero] = bit1Ptr[row];
				}

			}
			else{
				nonZero++; /*count the number of rows with non-zero entries in col*/
				nonZeroPtr[nonZero] = row;
			}

		} /*for(row)*/
		/*end copy rows with zeros in col to next tableau*/

		t2nrows = zero; /*current number of rows in mat2Ptr*/

		/*for the current column col*/
		for( row = 1; row <= nonZero; row++ ){	
			nzRow = nonZeroPtr[row]; /*non zero row number in matrix*/	
			
			/*find column in sparse row vector, corresponding to col*******************************************/
			STOP = my_false;
			countNz = 0;
			countCol = 1;

			for( xtype = 1; ((xtype <= NUMBER_OF_XTYPES) && (STOP == my_false)); xtype++ ){
				testBit = HIGH_BIT;
				xWord = bit1Ptr[nzRow][xtype];
				for( bits = 1; bits <= BIT_TYPESIZE; bits++ ){			
					if( (testBit & xWord) != 0 ){
						countNz++;
					}
					testBit = testBit >> 1;
					countCol++;
					if( countCol > col ){
						STOP = my_true;
						break;
					}
				}
			}
			if( STOP == my_false ){
				testBit = HIGH_BIT;
				xWord = bit1Ptr[nzRow][xtype];
				for( bits = 1;bits <= NUMBER_OF_RESIDS; bits++ ){		
					if( (testBit & xWord) != 0 ){
						countNz++;
					}
					testBit = testBit >> 1;
					countCol++;
					if( countCol > col ){
						break;
					}
				}
			}

			/*end find column*******************************************************************************/

			nzCol = countNz;
			/*find rows with negative entries*/
			if( mat1Ptr[nzRow][nzCol] < 0 ){
				negPtrRow[neg + 1] = nzRow;
				negPtrCol[neg + 1] = nzCol;
				neg++;
			}
			/*find rows with positive entries*/
			else{ /*else > 0*/
				posPtrRow[pos + 1] = nzRow;
				posPtrCol[pos + 1] = nzCol;
				pos++;
			}

		} /*for(row)*/
		npos = pos;
		nneg = neg;
		cnewRows = npos * nneg; /*potential number of rows to be added to the next tableau*/
		nnewRows = cnewRows + t2nrows; 

		/*reduce rows*/
		reduceRows(mat1Ptr, bit1Ptr, T1nrows);


		if( (nneg < npos) && (cnewRows > 0) ){
			tempNeg = nneg;
			nneg = npos;
			npos = tempNeg;
			tPosNegPtr = negPtrRow;
			negPtrRow = posPtrRow;
			posPtrRow = tPosNegPtr;
			tPosNegPtr = negPtrCol;
			negPtrCol = posPtrCol;
			posPtrCol = tPosNegPtr;
		}

		if( verbose == my_true ){
			columnNumber++;
			printf("%d. %d  %d \n", columnNumber, (col - NCOLS), nnewRows);
			fflush(stdout);
		}


		if( (nnewRows  > (t2nallRows - 1)) && (nnewRows <= MEMORY_THRESHOLD) ){/*expand next matrix if needed*/
			t2nallRows = nnewRows; 
			mat2Ptr = (XTYPE**) realloc( (void*) mat2Ptr, (unsigned) (t2nallRows + 1) * sizeof( XTYPE* ) );
			if( mat2Ptr == NULL ){
				nrerror("Failed to allocate pointer: mat2Ptr");
			}
			
			bit2Ptr = (UINT32**) realloc( (void*) bit2Ptr, (unsigned) (t2nallRows + 1) * sizeof( UINT32* ) );
			if( bit2Ptr == NULL ){
				nrerror("Failed to allocate pointer: bit2Ptr");
			}
		}
		else if( nnewRows > t2nallRows - 1 ){/*nnewRows > MEMORY_THRESHOLD*/
			if( t2nallRows < INITIAL_MEMORY ){
				t2nallRows = INITIAL_MEMORY;
			}
			else{
				t2nallRows += MEMORY_INCREMENT;
			}
			mat2Ptr = (XTYPE**) realloc( (void*) mat2Ptr, (unsigned) (t2nallRows) * sizeof( XTYPE* ) );
			if( mat2Ptr == NULL ){
				nrerror("Failed to allocate pointer: mat2Ptr");
			}
			
			bit2Ptr = (UINT32**) realloc( (void*) bit2Ptr, (unsigned) (t2nallRows) * sizeof( UINT32* ) );
			if( bit2Ptr == NULL ){
				nrerror("Failed to allocate pointer: bit2Ptr");
			}
			memAllocDummy = my_true;
		}


		/*newRowsBitPtr holds the bit reps and newRowsPosNegPtr holds the positive/negative rowindices */
		if( nnewAllocatedRows < nneg ){
			nnewAllocatedRows += nneg;
			if( (newRowsBitPtr = (BIT_TYPE**) realloc( newRowsBitPtr, (unsigned) 
							(nnewAllocatedRows + 1) * sizeof(BIT_TYPE*))) == NULL ){
				fprintf(stderr, "cannot reallocate newRowsBitPtr\n");
				exit(1);
			}
			if( (newRowsPosNegPtr = (UINT32**) realloc( newRowsPosNegPtr,
							(unsigned) (nnewAllocatedRows + 1) * sizeof(UINT32*))) == NULL ){
				fprintf(stderr, "cannot allocate memory for newRowsPosNegPtr\n");
				exit(1);
			}
		}

		numberOfRows = 0;
		numberOfTempRows = 0;


		/*temporary holder for bit rows*/
		if( nAllocatedTempRows < nneg ){
			nAllocatedTempRows += nneg;
			if( (tempBitPtr = (BIT_TYPE**) realloc( tempBitPtr, 
							(unsigned) (nAllocatedTempRows + 1) * sizeof(BIT_TYPE*))) == NULL ){
				fprintf(stderr, "cannot reallocate tempBitPtr\n");
				exit(1);
			}
			/*temporary holder for pos-neg rows*/
			if( (tempPosNegPtr = (UINT32**) realloc( tempPosNegPtr, 
							(unsigned) (nAllocatedTempRows + 1) * sizeof(UINT32*))) == NULL ){
				fprintf(stderr, "cannot allocate memory for tempPosNegPtr\n");
				exit(1);
			}
		}


		if( cnewRows > 0 ){/*add new rows to matrix*/
	        for( pos = 1; pos <= npos + 1; pos++ ){

			/*check independence of rows returned from previous check***********************************************************/

			if( nAllocatedTempRows < numberOfTempRows + nneg ){
				nAllocatedTempRows += numberOfTempRows + nneg;
				if( (tempBitPtr = (BIT_TYPE**) realloc( tempBitPtr, 
							(unsigned) (nAllocatedTempRows + 1) * sizeof(BIT_TYPE*))) == NULL ){
					fprintf(stderr, "cannot allocate memory for tempBitPtr\n");
					exit(1);
				}
				if( (tempPosNegPtr = (UINT32**) realloc( tempPosNegPtr,
							(unsigned) (nAllocatedTempRows + 1) * sizeof(UINT32*))) == NULL ){
					fprintf(stderr, "cannot allocate memory for tempPosNegPtr\n");
					exit(1);
				}
			}

			/*the bit rows and the positive/negative rows from the previous check are in newRowsBitPtr, newRowsPosNegPtr,
			 * respectively, for each iteration 1, 2, ..., nneg. The bit rows (positive/negative rows) that pass 
			 * the independence test will be stored in tempBitPtr (tempPosNegPtr). 
			 * The number of rows in these arrays is numberOfTempRows*/

			/*check independence for stored rows*/
			addRow = my_true;
			for( row = 1; row <= numberOfRows; row++ ){
				nextRowBitPtr = newRowsBitPtr[row];
				nextPosNegPtr = newRowsPosNegPtr[row];
				if( nextRowBitPtr == NULL ){
					continue;
				}						
				/*check if A(nextRow) is a subset of A(row)*/
				addRow = my_true;
				for( row2 = 1; row2 <= numberOfTempRows; row2++ ){	
					addRow = my_true;
					if( tempBitPtr[row2] == NULL ){
						continue;
					}
					addRow = my_false;
					for( xtype = 1; xtype <= numberOfXtypes; xtype++ ){
						if( ( (~nextRowBitPtr[xtype]) & tempBitPtr[row2][xtype] )
								!= 0 ){
							addRow = my_true;
							break; /*A(nextRow) is not a subset of A(row2) */
						}
					}
					if( (addRow == my_false) && (numberOfResiduals > 0) ){
						if( ( (~nextRowBitPtr[xtype]  & tempBitPtr[row2][xtype]) 
									& mask ) != 0 ){
							addRow = my_true; /*A(nRow) is not a subset of A(row2) */
						}
					}
					if( addRow == my_false ){
						break;
					}
				}/*for(row2)*/

				/* check if A(row) is a subset of A(nextRow) */
				if( addRow == my_true ){
					delRow = my_false;
					for( row2 = 1; row2 <= numberOfTempRows; row2++ ){
						delRow = my_false;
						if( tempBitPtr[row2] == NULL ){
							continue;
						}
						delRow = my_true;
						for( xtype = 1; xtype <= numberOfXtypes; xtype++ ){
							if( ( (~tempBitPtr[row2][xtype]) 
										& (nextRowBitPtr[xtype]) ) != 0 ){
								delRow = my_false;
								break; /*A(row2) is not a subset of A(nRow)*/
							}
						}
						if( (delRow == my_true) && (numberOfResiduals > 0) ){
							if( ( ( (~tempBitPtr[row2][xtype]) 
									& (nextRowBitPtr[xtype]) ) & mask ) != 0 ){
									delRow = my_false;
									/*A(row2) is not a subset of A(nRow)*/
							}
						}
						if (delRow == my_true ){
							freebitTypeVector(tempBitPtr[row2]);
							tempBitPtr[row2] = NULL;
							freebitTypeVector(tempPosNegPtr[row2]);
							tempPosNegPtr[row2] = NULL;
						}
					} /*for(row2)*/
					numberOfTempRows++;
					tempBitPtr[numberOfTempRows] = nextRowBitPtr;
					tempPosNegPtr[numberOfTempRows] = nextPosNegPtr;
				} /*if(addRow)*/
			}/*for(row)*/

			/*get rid of NULLs*/
			if( numberOfTempRows > 0 ){
				row = 1;
				while(1){
					while( tempBitPtr[row] == NULL ){
						for( row2 = row; row2 < numberOfTempRows; row2++ ){
							tempBitPtr[row2] = tempBitPtr[row2 + 1];
							tempPosNegPtr[row2] = tempPosNegPtr[row2 + 1];
						}
						numberOfTempRows--;
						if( row >= numberOfTempRows){
							if( tempBitPtr[row] == NULL ){/*last row*/
								numberOfTempRows--;
							}
							break;
						}
					}
					row++;
					if( row >= numberOfTempRows){
						if( (row == numberOfTempRows) && (tempBitPtr[row] == NULL) ){/*last row*/
							numberOfTempRows--;
						}
						break;
					}

				} /*while(1)*/
			} /*if( numberOfTempRows )*/

			numberOfRows = 0;

			if( pos <= npos ){
			for( neg = 1; neg <= nneg; neg++ ){

				/*check new rows to be added to matrix for mutual independence**********************************/

				/*pointer to new bit row to be added to next bit matrix*/
				if( (newRowBitPtr = (BIT_TYPE*) malloc( (unsigned) (NUMBER_OF_XTYPES + 1) * sizeof(BIT_TYPE) ) ) == NULL ){
					fprintf(stderr, "cannot allocate memory for newRowBitPtr\n");
					exit(1);
				}
				newRowBitPtr--;
				for( xtype = 1; xtype <= NUMBER_OF_XTYPES + 1; xtype++ ){
					/*create a vector whose bits are set if the corresponding element
					 *in the new row is non-zero.*/
					newRowBitPtr[xtype] = bit1Ptr[posPtrRow[pos]][xtype] |
								bit1Ptr[negPtrRow[neg]][xtype]; /*this is OK for xtype <= NCOLS*/
				}

				/*test for conical independence************************************************************/

				/*check if A(newRow) is a subset of A(row)*/
				addRow = my_true; /*if t2nrows == 0, then want addRow to be my_true*/
				for( row = 1; row <= t2nrows; row++ ){
					addRow = my_false;
					if( bit2Ptr[row] == NULL ){
						continue;
					}
					for( xtype = 1; xtype <= numberOfXtypes; xtype++ ){
						if( ( (~newRowBitPtr[xtype]) & bit2Ptr[row][xtype] ) != 0 ){
							addRow = my_true;
							break; /*A(nRow) is not a subset of A(row) */
						}
					}
					if( (addRow == my_false) && (numberOfResiduals > 0) ){
						if( ( (~newRowBitPtr[xtype]  & bit2Ptr[row][xtype]) & mask ) != 0 ){
							addRow = my_true; /*A(nRow) is not a subset of A(row) */
						}
					}
					if( addRow == my_false ){
						break;
					}
				}

				/*check if A(newRow) is a subset of 'A of previously stored rows'*/
				if( (addRow == my_true) && (numberOfRows >= 1) ){
					/*check if A(newRow) is a subset of A(row)*/
					addRow = my_true; 
					for( row = 1; row <= numberOfRows; row++ ){
						addRow = my_false;
						if( newRowsBitPtr[row] == NULL ){
							continue;
						}
						for( xtype = 1; xtype <= numberOfXtypes; xtype++ ){
							if( ( (~newRowBitPtr[xtype]) & newRowsBitPtr[row][xtype] ) != 0 ){
								addRow = my_true;
								break; /*A(nRow) is not a subset of A(row) */
							}
						}
						if( (addRow == my_false) && (numberOfResiduals > 0) ){
							if( ( (~newRowBitPtr[xtype]  & newRowsBitPtr[row][xtype]) 
										& mask ) != 0 ){
								addRow = my_true; /*A(nRow) is not a subset of A(row) */
							}
						}
						if( addRow == my_false ){
							break;
						}
					}
				} /*if(addRow)*/
				
				/* check if A(row) is a subset of A(newRow) */
				if( addRow == my_true ){
					for( row = 1; row <= t2nrows; row++ ){
						delRow = my_true;
						if( bit2Ptr[row] == NULL ){
							continue;
						}
						for( xtype = 1; xtype <= numberOfXtypes; xtype++ ){
							if( ( (~bit2Ptr[row][xtype]) & newRowBitPtr[xtype] ) != 0 ){
								delRow = my_false;
								break; /*A(row) is not a subset of A(nRow)*/
							}
						}
						if( (delRow == my_true) && (numberOfResiduals > 0) ){
							if( ( ( (~bit2Ptr[row][xtype]) & newRowBitPtr[xtype] ) & mask ) != 0 ){
									delRow = my_false;
									/*A(row) is not a subset of A(nRow)*/
							}
						}
						if (delRow == my_true ){
							(mat2Ptr[row])++;
							free((void*) mat2Ptr[row]);
							mat2Ptr[row] = NULL;
							(bit2Ptr[row])++;
							free((void*) bit2Ptr[row]);
							bit2Ptr[row] = NULL;
						}
					} /*for(row)*/
				} /*if(addRow)*/
				/* check if 'A of the stored rows' are subsets of A(newRow) */
				if( (addRow == my_true) && (numberOfRows >= 1) ){
					for( row = 1; row <= numberOfRows; row++ ){
						delRow = my_true;
						if( newRowsBitPtr[row] == NULL ){
							continue;
						}
						for( xtype = 1; xtype <= numberOfXtypes; xtype++ ){
							if( ( (~newRowsBitPtr[row][xtype]) & newRowBitPtr[xtype] ) 
									!= 0 ){
								delRow = my_false;
								break; /*A(row) is not a subset of A(nRow)*/
							}
						}
						if( (delRow == my_true) && (numberOfResiduals > 0) ){
							if( ( ( (~newRowsBitPtr[row][xtype]) & newRowBitPtr[xtype] ) 
										& mask ) != 0 ){
									delRow = my_false;
									/*A(row) is not a subset of A(nRow)*/
							}
						}
						if ( delRow == my_true ){
							(newRowsBitPtr[row])++;
							free((void*) newRowsBitPtr[row]);
							newRowsBitPtr[row] = NULL;
							(newRowsPosNegPtr[row])++;
							free((void*) newRowsPosNegPtr[row]);
							newRowsPosNegPtr[row] = NULL;
						}
					} /*for(row)*/
				} /*if(addRow)*/

				/*end test for conical independence************************************************************/

				if( addRow == my_true ){
					numberOfRows++;
					newRowsBitPtr[numberOfRows] = newRowBitPtr;
					if( ((newRowsPosNegPtr[numberOfRows]) = 
								(BIT_TYPE*) malloc((unsigned) 2 * sizeof(BIT_TYPE))) == NULL){
						fprintf(stderr, "cannot allocate memory for newRowsPosNegPtr\n");
						exit(1);
					}
					(newRowsPosNegPtr[numberOfRows])--;
					newRowsPosNegPtr[numberOfRows][POS] = pos;
					newRowsPosNegPtr[numberOfRows][NEG] = neg;
				}/*if(addRow)*/
				else{
					newRowBitPtr++;
					free((void*) newRowBitPtr);
				}
			}/*for(neg)*/
			}/*if(pos <= npos)*/

			} /*for(pos)*/
		}/*if(cnewRows)*/

	if( numberOfTempRows + t2nrows > t2nallRows ){/*expand next matrix if needed*/
		t2nallRows += numberOfTempRows + t2nrows; 
		if( (mat2Ptr = (XTYPE**) realloc( (void*) mat2Ptr, (unsigned) (t2nallRows + 1) * sizeof( XTYPE* ))) == NULL ){
			nrerror("Failed to allocate pointer: mat2Ptr");
		}
		
		if( (bit2Ptr = (BIT_TYPE**) realloc( (void*) bit2Ptr, (unsigned) (t2nallRows + 1) * sizeof( BIT_TYPE* ))) == NULL ){
			nrerror("Failed to allocate pointer: bit2Ptr");
		}
	}

	/*add the rows to the matrices*/
	for( row = 1; row <= numberOfTempRows; row++ ){
		if( tempBitPtr[row] == NULL ){
			continue;
		}
		t2nrows++;
		bit2Ptr[t2nrows] = tempBitPtr[row];
		pos = tempPosNegPtr[row][POS];
		neg = tempPosNegPtr[row][NEG];

		/*make sparse row******************************************************/
		nz = 0;
		highBit = HIGH_BIT;
		lowBit = LOW_BIT;
		for( xtype = 1; xtype <= NUMBER_OF_XTYPES; xtype++ ){
			tempBit = tempBitPtr[row][xtype];
			for( bits = 0; bits < BIT_TYPESIZE; bits++ ){
				if( (tempBit & lowBit) != 0 ){
					nz++;
				}
				tempBit = tempBit>>1;
			}
		}
		tempBit = tempBitPtr[row][xtype];
		for( bits = 0; bits < NUMBER_OF_RESIDS; bits++){
			if( (tempBit & highBit) != 0 ){
				nz++;
			}
			tempBit = tempBit<<1;
		}
		/*end find size -- size = nz*/

		/*pointer to new row to be added to next matrix*/
		newRowPtr = vector(nz);

		/*combine rows*************************************************/
		negRow = negPtrRow[neg];
		negCol = negPtrCol[neg];
		posRow = posPtrRow[pos];
		posCol = posPtrCol[pos];
		
		/*coefficient from the 'negative' row*/
		negCoeff = labs( mat1Ptr[negRow][negCol] );
		/*coefficient from the 'positive row'*/
		posCoeff = labs( mat1Ptr[posRow][posCol] );
		newCol = 1;
		posSlot = 1;
		negSlot = 1;
		bitCount = 0;
		STOP = my_false;
		/*find the least significant bit in tempBitPtr[row], i.e., the length of the new row */
		if( NUMBER_OF_RESIDS > 0 ){
			testBit = tempBitPtr[row][NUMBER_OF_XTYPES + 1];
		}
		else{
			testBit = tempBitPtr[row][NUMBER_OF_XTYPES];
		}
		for( bits = 1; bits <= BIT_TYPESIZE; bits++ ){
			if( (testBit & lowBit) != 0 ){
				break;
			}
			else{
				lowBit = lowBit << 1;
			}
		}
		if( NUMBER_OF_RESIDS > 0 ){
			bitStop =  NUMBER_OF_XTYPES * BIT_TYPESIZE + (BIT_TYPESIZE - bits + 1);
		}
		else{
			bitStop = (NUMBER_OF_XTYPES - 1) * BIT_TYPESIZE  + (BIT_TYPESIZE - bits + 1);
		}
		/*end find the least significant bit -- result is bitStop **********************************/

		/*combine the rows*/
		for( xtype = 1; xtype <= NUMBER_OF_XTYPES + 1; xtype++ ){
			nextBit = HIGH_BIT;
			if( STOP == my_true ){								
				break;
			}
			for( bits = 1; bits <= BIT_TYPESIZE; bits++ ){
				bitCount++; /*keep track of bits > NCOLS (we don't know if these entries
					      *are nonzero or zero since there are negative elements in
					      *columns > NCOLS*/
				if( bitCount > bitStop ){
					STOP = my_true;
					break;
				}
				if( ( tempBitPtr[row][xtype] & nextBit ) != 0 ){/*this means that either posRow or 
									*negRow has a non-zero entry 
									*in this column*/
					posBit = bit1Ptr[posRow][xtype] & nextBit;
					negBit = bit1Ptr[negRow][xtype] & nextBit;
					if( (posBit & negBit) != 0 ){
						newRowPtr[newCol] = negCoeff * mat1Ptr[posRow][posSlot] +
							posCoeff * mat1Ptr[negRow][negSlot];
						
						if( (bitCount > NCOLS) && (newRowPtr[newCol] == 0) ){
							tempBitPtr[row][xtype] = tempBitPtr[row][xtype] & (~nextBit); 
													/*clear the bit*/
						}
						else{
							newCol++;
						}
						posSlot++;
						negSlot++;
					}
					else{
						
						if( posBit != 0 ){
							newRowPtr[newCol] = negCoeff * 
								mat1Ptr[posRow][posSlot];
							newCol++;
							posSlot++;
						}
						else{
							if( negBit != 0 ){
								newRowPtr[newCol] = posCoeff *
									mat1Ptr[negRow][negSlot];
								newCol++;
								negSlot++;
							}
							else{
								fprintf(stderr,"Error in bitArray \n");
							}
						}
					}/*else*/
				} /*if(tempBitPtr[row])*/
				nextBit = nextBit >> 1;
			}/*for(bits)*/
		}/*for(xtype)*/
		/*end make sparse row*************************************************/

		mat2Ptr[t2nrows] = newRowPtr;
	} /*for(row)*/


		for( pos = 1; pos <= npos; pos++ ){
			freeVector( mat1Ptr[posPtrRow[pos]] );
			freebitTypeVector( bit1Ptr[posPtrRow[pos]] );
		}
		for( neg = 1; neg <= nneg; neg++ ){
			freeVector( mat1Ptr[negPtrRow[neg]] );
			freebitTypeVector( bit1Ptr[negPtrRow[neg]] );
		}

		if( stopDummy == my_true ){
			if( col == (stopColumn + NCOLS) ){
				if( verbose == my_true ){
					time(&endTime);
					printf("Exiting... time in seconds: %ld \n", (endTime - startTime));
				}
				writeMat2(mat2Ptr, bit2Ptr, t2nrows);
				exit(0);
			}
		}

		if( partialDummy == my_true ){
			nColsProcessed++;
			if( nColsProcessed == nCols2Process ){
				if( verbose == my_true ){
					time(&endTime);
					printf("Exiting... time in seconds: %ld \n", (endTime - startTime));
				}
				writeMat2(mat2Ptr, bit2Ptr, t2nrows);
				exit(0);
			}
		}
		
		T1nrows = t2nrows; 
		/*change pointers for next iteration*/
		tempRows = t1nallRows;
		t1nallRows = t2nallRows;
		t2nallRows = tempRows;

		tempPtr = mat1Ptr;
		tBitPtr = bit1Ptr;
		mat1Ptr = mat2Ptr;
		bit1Ptr = bit2Ptr;
		mat2Ptr = tempPtr;
		bit2Ptr = tBitPtr;
		

	} /*end main loop*/

	

	/** add TE ***********************************************************************/
	if( matEPtr != NULL ){
		for( u = 1; u <= uSize; u++ ){
			col = uxflxPtr[u];
			for( row = 1; row <= TEnrows; row++ ){
				if( matEPtr[row][col] != 0 ){
					TErow = row;
					break;
				}
			}
			/*find column in bit array*/
			colSlot = (col - 1)/BIT_TYPESIZE  + 1;
			if( (col - 1) % BIT_TYPESIZE  != 0 ){
	       		colBit =  HIGH_BIT >> ( (col - 1) % BIT_TYPESIZE );
			}
			else{
				colBit = HIGH_BIT;
			}
			for( row = 1; row <= T1nrows; row++ ){	
				if( bit1Ptr[row] == NULL ){
					continue;
				}				
				if( (bit1Ptr[row][colSlot] & colBit) != 0 ){
					newRowPtr = vector(NCOLS + NROWS);

			/*find column in sparse row vector, corresponding to col*******************************************/
			STOP = my_false;
			countNz = 0;
			countCol = 1;

			for( xtype = 1; ((xtype <= NUMBER_OF_XTYPES) && (STOP == my_false)); xtype++ ){
				testBit = HIGH_BIT;
				xWord = bit1Ptr[row][xtype];
				for( bits = 1; bits <= BIT_TYPESIZE; bits++ ){			
					if( (testBit & xWord) != 0 ){
						countNz++;
					}
					testBit = testBit >> 1;
					countCol++;
					if( countCol > col ){
						STOP = my_true;
						break;
					}
				}
			}
			if( STOP == my_false ){
				testBit = HIGH_BIT;
				xWord = bit1Ptr[row][xtype];
				for( bits = 1; bits <= NUMBER_OF_RESIDS; bits++ ){		
					if( (testBit & xWord) != 0 ){
						countNz++;
					}
					testBit = testBit >> 1;
					countCol++;
					if( countCol > col ){
						break;
					}
				}
			}

			/*end find column*******************************************************************************/

					T1coeff = mat1Ptr[row][countNz];
					xtype = 1;
					testBit = HIGH_BIT;
					mat1Slot = 1;
					newSlot = 1;
					for( TEcol = 1; TEcol <= NROWS + NCOLS; TEcol++ ){
						matEValue = T1coeff * matEPtr[TErow][TEcol];
						if( (bit1Ptr[row][xtype] & testBit) != 0){
							if( (newRowPtr[newSlot] = matEValue + mat1Ptr[row][mat1Slot]) != 0 ){
								newSlot++;
							}
							else{
								bit1Ptr[row][xtype] = bit1Ptr[row][xtype] & (~testBit); /*unset bit*/
							}
							mat1Slot++;							
						}
						else{
							if( matEValue != 0 ){
								newRowPtr[newSlot] = matEValue;
								newSlot++;
								bit1Ptr[row][xtype] = bit1Ptr[row][xtype] | testBit; /*set the bit*/
							}
						}
						testBit = testBit >> 1;
						if( (TEcol % BIT_TYPESIZE) == 0 ){
							xtype++;
							testBit = HIGH_BIT;
						}
					}
					/*find the size of the new row*/
					nz = 0;
					highBit = HIGH_BIT;
					lowBit = BIT_1;
					for( xtype = 1; xtype <= NUMBER_OF_XTYPES; xtype++ ){
						tempBit = bit1Ptr[row][xtype];
						for( bits = 0; bits < BIT_TYPESIZE; bits++ ){
							if( (tempBit & lowBit) != 0x0 ){
								nz++;
							}
							tempBit = tempBit>>1;
						}
					}
					tempBit = bit1Ptr[row][xtype];
					for( bits = 0; bits < NUMBER_OF_RESIDS; bits++){
						if( (tempBit & highBit) != 0x0 ){
							nz++;
						}
						tempBit = tempBit<<1;
					}
					/*pointer to new row to be added to next matrix*/
					
					vectorSize = nz * sizeof(XTYPE);
					newRowPtr++; /*need to do this when using realloc to compensate for malloc's return value*/
					newRowPtr = (XTYPE*) realloc( (void*) newRowPtr, (unsigned) vectorSize);
					newRowPtr--;
					freeVector(mat1Ptr[row]);
					mat1Ptr[row] = newRowPtr;
				}

								
			}/*for(row)*/
		} /*for(u)*/
	} /*if(matEPtr)*/
	/* end add TE **************************************************************************************/
	
	freeuint32Vector(uxflxPtr);
	freeuint32Vector(posPtrRow);
	freeuint32Vector(negPtrRow);
	freeuint32Vector(posPtrCol);
	freeuint32Vector(negPtrCol);
	sortedPtrs++;
	free((void*) sortedPtrs);
	if( newColumns != NULL ){
		free((void*) newColumns);
	}
	

	*mat1PtrPtr = mat1Ptr;
	*mat2PtrPtr = mat2Ptr;
	*bit1PtrPtr = bit1Ptr;

} /*computePathways()*/

/***************************************************************************************************************************************
 *
 * 	Function: 	readMatrix()
 *
 * 	Description:	This function constructs the initial tableau and its corresponding bitmap from the input files containing the
 * 			stoichiometric matrix and the values of the exchange fluxes.
 *
 * 	Parameters:	m1Ptr -- the matrix which will contain the initial tableau (minus the rows corresponding to the unconstrained
 * 			exchange fluxes) after the input files have been read
 * 			matEPtr -- the matrix which will contain the rows corresponding to the unconstrained exchange fluxes (TE)
 * 			bit1Ptr -- the matrix which will contain the bit map corresponding to m1Ptr
 * 			xPtr -- an array containing the values of the exchange fluxes (input = -1, free = 0, output = 1)
 * 			file -- the file name of the file containing the stoichiometric matrix
 * 			xFile -- the file name of the file containing the values of the exchange fluxes
 *
 *	Date:		April 24, 2004
 *
 * 	Author:		Steve Bell
 *
 * ***************************************************************************************************************************************/

void readMatrix(XTYPE** m1Ptr,  XTYPE** matEPtr, BIT_TYPE** bit1Ptr,
							XFLXTYPE* xPtr, char* file, char* xFile)
{
	FILE *fp, *xfp;
        XTYPE inXtype, xtype;
	XTYPE **mat1Ptr;
		XFLXTYPE inXFLX;
	UINT32 zcol, row = 1, trow = 1, terow = 1, col = NCOLS, xflx = 1, nFREE = 0;
	UINT32 nzCol, vectorSize;
	INT32 scanReturn;

	XFLXTYPE xflxType = FREE;

	mat1Ptr = xtypeMatrix(NCOLS, NROWS + NCOLS);
	

	/*input the exchange flux vector*/
	if( NXFLX > 0 ){
		if( ( xfp = fopen(xFile, "r") ) == NULL ){
			fprintf(stderr, "unable to open file %s", xFile);
			exit(3);
		}
		else{
			while( fscanf( xfp, "%lu", &inXFLX ) != EOF ){
				xPtr[xflx] =  inXFLX;
				if( inXFLX == FREE ){
					nFREE++;
				}
				xflx++;
			}
		}
		if( xflx > NXFLX + 1 ){
			fprintf(stderr, "error in input file %s for exchange fluxes \n", xFile);
			fclose(xfp);
			exit(4);
		}

			if( nFREE == 0 ){
			matEPtr = NULL;
 		}	
		fclose(xfp);
	}

    if( ( fp = fopen(file, "r") ) == NULL ){
		fprintf(stderr, "unable to open file %s", file);
		exit(3);
	}
    else{
	/*input the transpose of the stoichiometrix matrix*/		
		for( col = NCOLS + 1; col <= NCOLS + NROWS; col++ ){
			trow = XFLXSTART;
			terow = 1;
			for( row = 1; row <= NCOLS; row++ ){
				if( (scanReturn = fscanf(fp, "%d", &inXtype) ) == EOF ){				
					printf("Possible error in input argument -- too few rows?\n");
					break; /*should not get here*/
				}
				if( scanReturn == 0 ){
						fprintf(stderr, "error in input matrix -- possibly row %ld\n", (col - NCOLS));
						exit(2);
				}
				if( row <= XFLXSTART - 1){
					if( col == NCOLS + 1){
						for( zcol = 1; zcol <= NCOLS; zcol++ ){/*identity matrix*/
							if( zcol == row ){
								mat1Ptr[row][zcol] = 1; 
							}
							else{
								mat1Ptr[row][zcol] = 0;
							}
						}
					}
					mat1Ptr[row][col] = inXtype;
				}
				else{
					xflxType = xPtr[ row - XFLXSTART + 1 ]; 
					if( xflxType == FREE ){
						if( col == NCOLS + 1 ){/*identity matrix*/
							for( zcol = 1; zcol <= NCOLS; zcol++ ){
								if( zcol == row ){
									matEPtr[terow][zcol] = 1;
								}
								else{
									matEPtr[terow][zcol] = 0;
								}
							}
						}
						matEPtr[terow][col] = inXtype;
						terow++;
					}
					else{
						if( xflxType == INPUT ){
							if( col == NCOLS + 1 ){
								for( zcol = 1; zcol <= NCOLS; zcol++ ){
									if( zcol == row ){/*negative identity row*/
								
										mat1Ptr[trow][zcol] = -1;
									}
									else{
										mat1Ptr[trow][zcol] = 0;
									}
								}
							}
							
							mat1Ptr[trow][col] = - inXtype; /*mult trow by -1*/
							trow++;
						}
						else{
							if( col == NCOLS + 1){
								for( zcol = 1; zcol <= NCOLS; zcol++ ){
									if( zcol == row ){
								
										mat1Ptr[trow][zcol] = 1;
									}
									else{
										mat1Ptr[trow][zcol] = 0;
									}
								}
							}
							
							mat1Ptr[trow][col] = inXtype;
							trow++;
						}
					} /*else( xflxType != FREE )*/
				} /*else( row >= XFLXSTART )*/
			} /*for(row)*/
		}/*for(col)*/
	}
	TEnrows = terow - 1;
    T1nrows = trow - 1;
		

	if( (col > NROWS + NCOLS  + 1) || (TEnrows + T1nrows != NCOLS) ){
		fprintf(stderr, "error in input file %s", file);
		fclose(fp);
		exit(4);
	}
	fclose(fp);

	for( row = 1; row <= T1nrows; row++ ){
		makeBitArray( mat1Ptr[row], bit1Ptr[row]);
		vectorSize = numberOfNz( bit1Ptr[row] );
		nzCol = 1;
		m1Ptr[row] = vector( vectorSize );
		for( col = 1; col <= NCOLS + NROWS; col++ ){
			xtype = mat1Ptr[row][col];
			if( xtype != 0 ){
				m1Ptr[row][nzCol] = xtype;
				nzCol++;
			}
		}
	}
	freextypeMatrix( mat1Ptr, T1nrows );
} /*readMatrix()*/

/***************************************************************************************************************************************
 *
 * 	Function: 	readMatrix2()
 *
 * 	Description:	This function constructs the initial tableau and its corresponding bitmap from the input reaction file (.expa or
 * 			.cel).
 * 			
 *
 * 	Parameters:	m1Ptr -- the matrix which will contain the initial tableau (minus the rows corresponding to the unconstrained
 * 			exchange fluxes) after the input files have been read
 * 			matEPtr -- the matrix which will contain the rows corresponding to the unconstrained exchange fluxes (TE)
 * 			bit1Ptr -- the matrix which will contain the bit map corresponding to m1Ptr
 * 			xPtr -- an array containing the values of the exchange fluxes (input = -1, free = 0, output = 1)
 * 			stoickMat -- pointer to the matrix containing the stoichiometric matrix 
 *
 *	Date:		April 24, 2004
 *
 * 	Author:		Steve Bell
 *
 * ***************************************************************************************************************************************/

void readMatrix2(XTYPE** m1Ptr,  XTYPE** matEPtr, BIT_TYPE** bit1Ptr,
							XFLXTYPE* xPtr, XTYPE*** stoickMat)
{
        XTYPE xtype;
	XTYPE **mat1Ptr, **sMatrix;
	UINT32 zcol, row, trow, terow, col, xflx;
	UINT32 nzCol, vectorSize, sCol;
	BOOLEAN free;

	XFLXTYPE xflxType = FREE;

	mat1Ptr = xtypeMatrix(NCOLS, NROWS + NCOLS);
	sMatrix = *stoickMat;
	free = my_false;
	for(xflx = 1; xflx <= NXFLX; xflx++){
		if( xPtr[xflx] == FREE ){
			free = my_true;
			break;
		}
	}
	if( free == my_false ){
		matEPtr = NULL;
 	}

	
	trow = 1;
	terow = 1;	
	/*input the transpose of the stoichiometrix matrix*/		
	for( col = NCOLS + 1; col <= NCOLS + NROWS; col++ ){
		trow = XFLXSTART;
		terow = 1;
		sCol = col - NCOLS;
		for( row = 1; row <= NCOLS; row++ ){
			if( row <= XFLXSTART - 1){
				if( col == NCOLS + 1){
					for( zcol = 1; zcol <= NCOLS; zcol++ ){/*identity matrix*/
						if( zcol == row ){
							mat1Ptr[row][zcol] = 1; 
						}
						else{
							mat1Ptr[row][zcol] = 0;
						}
					}
				}
				mat1Ptr[row][col] = sMatrix[sCol][row];
			}
			else{
				xflxType = xPtr[ row - XFLXSTART + 1 ]; 
				if( xflxType == FREE ){
					if( col == NCOLS + 1 ){/*identity matrix*/
						for( zcol = 1; zcol <= NCOLS; zcol++ ){
							if( zcol == row ){
								matEPtr[terow][zcol] = 1;
							}
							else{
								matEPtr[terow][zcol] = 0;
							}
						}
					}
					matEPtr[terow][col] = sMatrix[sCol][row];
					terow++;
				}
				else{
					if( xflxType == INPUT ){
						if( col == NCOLS + 1 ){
							for( zcol = 1; zcol <= NCOLS; zcol++ ){
								if( zcol == row ){/*negative identity row*/
									mat1Ptr[trow][zcol] = -1;
								}
								else{
									mat1Ptr[trow][zcol] = 0;
								}
							}
						}
						
						mat1Ptr[trow][col] = - sMatrix[sCol][row]; /*mult trow by -1*/
						trow++;
					}
					else{
						if( col == NCOLS + 1){
							for( zcol = 1; zcol <= NCOLS; zcol++ ){
								if( zcol == row ){
							
									mat1Ptr[trow][zcol] = 1;
								}
								else{
									mat1Ptr[trow][zcol] = 0;
								}
							}
						}
						
						mat1Ptr[trow][col] = sMatrix[sCol][row];
						trow++;
					}
				} /*else( xflxType != FREE )*/
			} /*else( row >= XFLXSTART )*/
		} /*for(row)*/
	}/*for(col)*/

	TEnrows = terow - 1;
	T1nrows = trow - 1;
	
		

	if( (col > NROWS + NCOLS  + 1) || (TEnrows + T1nrows != NCOLS) ){
		fprintf(stderr, "error in input file ");
		exit(4);
	}


	for( row = 1; row <= T1nrows; row++ ){
		makeBitArray( mat1Ptr[row], bit1Ptr[row]);
		vectorSize = numberOfNz( bit1Ptr[row] );
		nzCol = 1;
		m1Ptr[row] = vector( vectorSize );
		for( col = 1; col <= NCOLS + NROWS; col++ ){
			xtype = mat1Ptr[row][col];
			if( xtype != 0 ){
				m1Ptr[row][nzCol] = xtype;
				nzCol++;
			}
		}
	}
	freextypeMatrix( mat1Ptr, T1nrows );
} /*readMatrix2()*/

/***************************************************************************************************************************************
 *
 * 	Function: 	readMatrix3()
 *
 * 	Description:	This function reads the imported matrix and conststructs the sparse matrix and the bit array.
 * 			It also constructs a pointer to the column indices that still need to be processed (the non-zero
 * 			columns from the right-hand side of the imported matrix
 *
 * 	Parameters:	m1Ptr -- pointer to the sparse matrix
 * 			bit1Ptr -- pointer to the bit array
 * 			matFile -- ascii file containing the imported matrix
 *
 *	Date:		January 25, 2005
 *
 * 	Author:		Steve Bell
 *
 * ***************************************************************************************************************************************/

void readMatrix3(XTYPE*** m1PtrPtr, BIT_TYPE*** bit1PtrPtr, char* matFile)
{
	FILE *fpMat;
	UINT32 row, col;
	UINT32 nzCol, vectorSize;
	XTYPE inXtype, xtype;
	XTYPE** mat1Ptr;
	XTYPE** m1Ptr;
	INT32 scanReturn;
	UINT32 nRowsAllocated = T1nrows;
	UINT32 rowIncrement = 500;
	BIT_TYPE** bit1Ptr;

	if( ( mat1Ptr = (XTYPE**) malloc( (unsigned) (nRowsAllocated + 1) * sizeof(XTYPE*) ) )== NULL ){
		fprintf(stderr, "cannot allocate mat1Ptr\n");
		exit(1);
	}

	if( ( bit1Ptr = (BIT_TYPE**) malloc( (unsigned) (nRowsAllocated + 1) * sizeof(BIT_TYPE*) ) ) == NULL ){
		fprintf(stderr, "cannot allocate bit1Ptr\n");
		exit(1);
	}

	if( ( fpMat = fopen(matFile, "r") ) == NULL ){
		fprintf(stderr, "unable to open file %s", matFile);
		exit(3);
	}

	row = 0;	
	/*input the matrix*/
	while( ( scanReturn = fscanf(fpMat, "%ld", &inXtype) ) != EOF ){
		if( scanReturn == 0 ){
			fprintf(stderr, "error in input matrix\n");
			exit(3);
		}
		row++;
		if( row > nRowsAllocated ){
			nRowsAllocated += rowIncrement;
			if( ( mat1Ptr = (XTYPE**) realloc(mat1Ptr, (unsigned) (nRowsAllocated + 1) * sizeof(XTYPE*) ) ) == NULL ){
				fprintf(stderr, "cannot reallocate m1Ptr\n");
				exit(1);
			}

			if( ( bit1Ptr = (BIT_TYPE**) realloc(bit1Ptr, (unsigned) (nRowsAllocated + 1) * sizeof(BIT_TYPE*) ) ) == NULL ){
				fprintf(stderr, "cannot reallocate bit1Ptr\n");
				exit(1);
			}
		}
		mat1Ptr[row] = vector(NROWS + NCOLS); 
		bit1Ptr[row] = bitTypeVector(NUMBER_OF_XTYPES + 1);
		col = 1;
		mat1Ptr[row][col] = inXtype;
		for( col = 2; col <= (NROWS + NCOLS); col++ ){
			scanReturn = fscanf(fpMat, "%ld", &inXtype); 
			if( scanReturn == 0 ){
				fprintf(stderr, "error in input matrix: (%ld, %ld)\n, row, col");
				exit(3);
			}
			mat1Ptr[row][col] = inXtype;
		} /*for(col)*/
	}/*while(scanReturn)*/
	T1nrows = row;

	if( ( m1Ptr = (XTYPE**) malloc( (unsigned) (T1nrows + 1) * sizeof(XTYPE*) ) )== NULL ){
		fprintf(stderr, "cannot allocate m1Ptr\n");
		exit(1);
	}
	
	for( row = 1; row <= T1nrows; row++ ){
		makeBitArray( mat1Ptr[row], bit1Ptr[row]);
		if( (vectorSize = numberOfNz( bit1Ptr[row] )) == 0 ){
				bit1Ptr[row] = NULL;
				m1Ptr[row] = NULL;
				continue;
		}
		nzCol = 1;
		m1Ptr[row] = vector( vectorSize );
		for( col = 1; col <= NROWS + NCOLS; col++ ){
			xtype = mat1Ptr[row][col];
			if( xtype != 0 ){
				m1Ptr[row][nzCol] = xtype;
				nzCol++;
			}
		}
	}
	for( row = 1; row <= T1nrows; row++ ){
		freeVector(mat1Ptr[row]);
	}
	*m1PtrPtr = m1Ptr;
	*bit1PtrPtr = bit1Ptr;
	free((void*) mat1Ptr);
	fclose(fpMat);

} /*readMatrix3()*/


/***************************************************************************************************************************************
 *
 * 	Function:	makeColPtr()
 *
 * 	Description:	Function makes pointer to user defined columns to process
 *
 * 	Parameters:	
 * 			colPtr (OUT) -- pointer to the columns which need to be processed
 * 			colFile -- ascii file containing the columns which to be processed
 * 			numberOfCols (OUT) -- number of columns in colFile
 * 			
 *
 * 	Date:		April 24, 2004
 *
 * 	Author:		Steve Bell
 *
 ***************************************************************************************************************************************/
void makeColPtr(UINT32* colPtr, char* colFile, UINT32* numberOfCols)
{
	FILE *fpCol;
	UINT32 inCol, nCols; 
	INT32 scanReturn;

	if( ( fpCol = fopen(colFile, "r") ) == NULL ){
		fprintf(stderr, "unable to open file %s", colFile);
		exit(3);
	}
	/*input columns*/
	nCols = 0;
	while( scanReturn = fscanf(fpCol, "%ld", &inCol) != EOF ){
		if( scanReturn == 0 ){
			fprintf(stderr, "error in column matrix\n");
			exit(2);
		}
		nCols++;
		colPtr[nCols] = (inCol + NCOLS);
	}
	*numberOfCols = nCols;
	fclose(fpCol);
}/*makeColPtr()*/

/***************************************************************************************************************************************
 *
 * 	Function:	numberOfNz()
 *
 * 	Description:	Function calculates the number of non-zero entries in a row of a bit map
 *
 * 	Parameters:	bPtr -- pointer to the row of the bit map
 *
 * 	Date:		April 24, 2004
 *
 * 	Author:		Steve Bell
 *
 ***************************************************************************************************************************************/


UINT32 numberOfNz( BIT_TYPE *bPtr)
{
	BIT_TYPE tempBit, highBit = HIGH_BIT, lowBit = LOW_BIT;
	UINT32 xtype, bits, nz;

	nz = 0;
	for( xtype = 1; xtype <= NUMBER_OF_XTYPES; xtype++ ){
		tempBit = bPtr[xtype];
		for( bits = 0; bits < BIT_TYPESIZE; bits++ ){
			if( (tempBit & lowBit) != 0x0 ){
				nz++;
			}
			tempBit = tempBit>>1;
		}
	}
	tempBit = bPtr[xtype];
	for( bits = 0; bits < NUMBER_OF_RESIDS; bits++){
		if( (tempBit & highBit) != 0x0 ){
			nz++;
		}
		tempBit = tempBit<<1;
	}
	
	return( nz );
}/*numberOfNz()*/


/***********************************************************************************************************************************
 *
 * 	Function:	makeBitArray()
 *
 * 	Description:	The function takes a row of the of the current tableau, converts it into bit form, and adds it to the
 * 			bit map.
 *
 * 	Parameters:	mPtr -- a pointer to the row of the current tableau
 * 			bArrayPtr -- a pointer to the row to be added to the bit map
 *
 * 	Date:		April 24, 2004
 *
 * 	Author:		Steve Bell
 *
 * *********************************************************************************************************************************/
			
void makeBitArray(XTYPE* mPtr, BIT_TYPE* bArrayPtr)
{	 
	BIT_TYPE nextBit, returnBit;	
	UINT32 xtype, xt, bits, startValue;
	XTYPE xTypeValue;
	
	for( xtype = 1; xtype <= NUMBER_OF_XTYPES; xtype++ ){
		returnBit = 0x0;
		nextBit = HIGH_BIT;
		startValue = (xtype - 1) * BIT_TYPESIZE;
		for( bits = 1; bits <= BIT_TYPESIZE; bits++ ){
			xTypeValue = mPtr[startValue + bits];
			if( xTypeValue != 0 ){
				returnBit |= nextBit;
			}
			nextBit = nextBit >> 1;
		}
		bArrayPtr[xtype] = returnBit;
	}
	nextBit = HIGH_BIT;
	returnBit = 0x0;
	if( NUMBER_OF_XTYPES == 0 ){
		xt = 0;
	}
	else{
		xt = (xtype - 1) * BIT_TYPESIZE;
	}
	for( bits = 1; bits <= NUMBER_OF_RESIDS; bits++ ){
		xTypeValue = mPtr[xt + bits];
		if( xTypeValue != 0 ){
			returnBit |= nextBit;
		}
		nextBit = nextBit >> 1;
	}
	bArrayPtr[xtype] = returnBit;
}/*makeBitArray()*/


void printHelp(void)
{
	printf("Functionality: \n");
	printf("1. Compute ExPas using a reaction file (.expa or .cel)\n");
	printf("   Command: 'expa file_name' (where 'file_name' is the name of the reaction file)\n");
	printf("   If the flag '-v' is used, then, for each iteration, the program will print\n");
	printf("      the column number (row number of the stoichiometric matrix) along with an upper \n");
	printf("      bound for the number of rows in the\n");
	printf("      next matrix (i.e., it prints the quantity (zero + pos * neg), where zero\n");
	printf("      is the number of rows whose entries in the current column are zero, pos is\n");
	printf("      the number of rows whose entries in the current column are positive, and\n");
	printf("      neg are the negative rows. \n");
	printf("   Example: 'expa -v file.expa'\n");
	printf("   The program will print out (in verbose mode) the extreme pathways to a file 'Paths.txt'\n");
	printf("   If the flag '-c' is used, the program will stop at a particular column (row of the\n");
	printf("   stoichiometric matrix)\n");
	printf("   Example: expa -cv file.expa 10\n");
	printf("   The program will stop at column 10 (row 10 of the S matrix) and \n");
	printf("   print out the path matrix and the remaining colums\n");
	printf("   to files 'mat.txt'and 'cols.txt', respectively.\n");
	printf("   If the flag '-o' is used, the program will process columns\n");
	printf("   in the order given by a file listing the columns\n");
	printf("   Example: 'expa -o file.expa col_file.txt'\n"); 
	printf("   The program will process the columns given in col_file.txt in order and\n");
	printf("   continue processing any remaining columns to the end of normal execution\n");
	printf("   If the flag '-p' is used, the program will process n colums\n");
	printf("   Example: 'expa -vp file.expa 5'\n");
	printf("   The program will process 5 columns and print the current path matrix to 'mat.txt' and \n");
       	printf("   the remaining colums to 'cols.txt' and exit.\n");
	printf("   If the flag '-r' is used, the program  will process columns given in a file and\n");
	printf("   begin from the paths given in the file 'mat.txt'.\n");
	printf("   Example: 'expa -r file.expa mat.txt cols.txt'\n");
    printf("   The program will process the colums given in cols.txt using the file 'mat.txt'\n");
	printf("   as the starting point of execution. The program will then continue normal execution.\n");
	printf("   The idea is to execute 'expa -c file.expa 10' and 'expa -r file.expa mat.txt cols.txt\n");
	printf("   in succession.\n");
	printf("   The flags can also be used in combination: For example\n");
	printf("   'expa -rp file.expa mat.txt cols.txt 2\n");
	printf("   The program will use the input files mat.txt and cols.txt generated by an earlier call\n");
	printf("   to expa and will then process the next two columns and exit (printing new files mat.txt and cols.txt)\n");
	printf("2. Output the stoichiometric matrix or the exchange fluxes.\n");
	printf("   Command: 'expa -s file_name' (where `file_name' is the name of the reaction file)\n");
	printf("   if 'expa -sv file_name' is used then all metabolites, reactions, and exchange\n");
	printf("       fluxes will be printed to standard output as well\n");
	printf("   (To redirect the output to a file, do 'expa -s file_name' > mat_file)\n");
	printf("   To output the exchange fluxes type 'expa -x file_name'\n");
	printf("3. Compute ExPas using the stoichiometric matrix as input\n");
	printf("   expa file1 file2 nrows, ncols, nxflx (where `file1' is the file containing the\n");
	printf("       stoichiometric matrix (a text file), `file2' is a file which contains the types\n");
	printf("       of the exchange fluxes, `nrows' is the number of rows of the matrix, `ncols' the\n");
	printf("       number of columns of the matrix, and `nxflx' is the number of exchange fluxes\n");
	printf("       The file 'file2' has one number in each row, indicating the type of the corresponding\n");
	printf("       exchange flux (-1 = input, 0 = free, 1 = output).\n");
	printf("       If there are no exchange fluxes (i.e., nxflx = 0), then the argument 'file2' is\n");
	printf("       the name of a file which need not exist (the argument needs to be present, though.\n");
	printf("       All of the functionality described for Functionality 1, can be applied here too.\n");
	printf("       For example, 'expa -vcr file1 file2 mat.txt cols.txt nrows ncols nxfluxes stopCol' will\n");
	printf("       use the input files mat.txt and cols.txt generated from an earlier call to expa\n");
	printf("       to process columns until column stopCol is reached and will then print out the matrix along\n");
	printf("       with new files mat.txt and cols.txt\n");
	printf("NOTE: the extreme pathways will be output to a file names `Paths.txt' where the rows\n");
	printf("      are the extreme pathways\n");
} /*printHelp()*/


/***************************************************************************************************************************************
 *
 * 	Function: 	reduceRows()
 *
 * 	Description:	This function reduces the components of a row by dividing each component by successive primes, if all
 * 			components are divisible by this prime.
 * 			
 *
 * 	Parameters:     matPtr -- the full matrix
 * 			bitPtr -- the bit array
 * 			numberOfRows -- number of rows in the matrices	
 *
 *	Date:		December 14, 2004
 *
 * 	Author:		Steve Bell
 *
 * ***************************************************************************************************************************************/

void reduceRows(XTYPE** matPtr, BIT_TYPE** bitPtr, UINT32 numberOfRows)
{
	XTYPE* matRow;
	UINT32 i, j, k, max, nextPrime, row, rowLength;
	UINT32 primes[1000] = {
2, 3, 5, 7, 11, 13, 17, 19, 23, 29, 
31, 37, 41, 43, 47, 53, 59, 61, 67, 71, 
73, 79, 83, 89, 97, 101, 103, 107, 109, 113, 
127, 131, 137, 139, 149, 151, 157, 163, 167, 173, 
179, 181, 191, 193, 197, 199, 211, 223, 227, 229, 
233, 239, 241, 251, 257, 263, 269, 271, 277, 281, 
283, 293, 307, 311, 313, 317, 331, 337, 347, 349, 
353, 359, 367, 373, 379, 383, 389, 397, 401, 409, 
419, 421, 431, 433, 439, 443, 449, 457, 461, 463, 
467, 479, 487, 491, 499, 503, 509, 521, 523, 541, 
547, 557, 563, 569, 571, 577, 587, 593, 599, 601, 
607, 613, 617, 619, 631, 641, 643, 647, 653, 659, 
661, 673, 677, 683, 691, 701, 709, 719, 727, 733, 
739, 743, 751, 757, 761, 769, 773, 787, 797, 809, 
811, 821, 823, 827, 829, 839, 853, 857, 859, 863, 
877, 881, 883, 887, 907, 911, 919, 929, 937, 941, 
947, 953, 967, 971, 977, 983, 991, 997, 1009, 1013, 
1019, 1021, 1031, 1033, 1039, 1049, 1051, 1061, 1063, 1069, 
1087, 1091, 1093, 1097, 1103, 1109, 1117, 1123, 1129, 1151, 
1153, 1163, 1171, 1181, 1187, 1193, 1201, 1213, 1217, 1223, 
1229, 1231, 1237, 1249, 1259, 1277, 1279, 1283, 1289, 1291, 
1297, 1301, 1303, 1307, 1319, 1321, 1327, 1361, 1367, 1373, 
1381, 1399, 1409, 1423, 1427, 1429, 1433, 1439, 1447, 1451, 
1453, 1459, 1471, 1481, 1483, 1487, 1489, 1493, 1499, 1511, 
1523, 1531, 1543, 1549, 1553, 1559, 1567, 1571, 1579, 1583, 
1597, 1601, 1607, 1609, 1613, 1619, 1621, 1627, 1637, 1657, 
1663, 1667, 1669, 1693, 1697, 1699, 1709, 1721, 1723, 1733, 
1741, 1747, 1753, 1759, 1777, 1783, 1787, 1789, 1801, 1811, 
1823, 1831, 1847, 1861, 1867, 1871, 1873, 1877, 1879, 1889, 
1901, 1907, 1913, 1931, 1933, 1949, 1951, 1973, 1979, 1987, 
1993, 1997, 1999, 2003, 2011, 2017, 2027, 2029, 2039, 2053, 
2063, 2069, 2081, 2083, 2087, 2089, 2099, 2111, 2113, 2129, 
2131, 2137, 2141, 2143, 2153, 2161, 2179, 2203, 2207, 2213, 
2221, 2237, 2239, 2243, 2251, 2267, 2269, 2273, 2281, 2287, 
2293, 2297, 2309, 2311, 2333, 2339, 2341, 2347, 2351, 2357, 
2371, 2377, 2381, 2383, 2389, 2393, 2399, 2411, 2417, 2423, 
2437, 2441, 2447, 2459, 2467, 2473, 2477, 2503, 2521, 2531, 
2539, 2543, 2549, 2551, 2557, 2579, 2591, 2593, 2609, 2617, 
2621, 2633, 2647, 2657, 2659, 2663, 2671, 2677, 2683, 2687, 
2689, 2693, 2699, 2707, 2711, 2713, 2719, 2729, 2731, 2741, 
2749, 2753, 2767, 2777, 2789, 2791, 2797, 2801, 2803, 2819, 
2833, 2837, 2843, 2851, 2857, 2861, 2879, 2887, 2897, 2903, 
2909, 2917, 2927, 2939, 2953, 2957, 2963, 2969, 2971, 2999, 
3001, 3011, 3019, 3023, 3037, 3041, 3049, 3061, 3067, 3079, 
3083, 3089, 3109, 3119, 3121, 3137, 3163, 3167, 3169, 3181, 
3187, 3191, 3203, 3209, 3217, 3221, 3229, 3251, 3253, 3257, 
3259, 3271, 3299, 3301, 3307, 3313, 3319, 3323, 3329, 3331, 
3343, 3347, 3359, 3361, 3371, 3373, 3389, 3391, 3407, 3413, 
3433, 3449, 3457, 3461, 3463, 3467, 3469, 3491, 3499, 3511, 
3517, 3527, 3529, 3533, 3539, 3541, 3547, 3557, 3559, 3571, 
3581, 3583, 3593, 3607, 3613, 3617, 3623, 3631, 3637, 3643, 
3659, 3671, 3673, 3677, 3691, 3697, 3701, 3709, 3719, 3727, 
3733, 3739, 3761, 3767, 3769, 3779, 3793, 3797, 3803, 3821, 
3823, 3833, 3847, 3851, 3853, 3863, 3877, 3881, 3889, 3907, 
3911, 3917, 3919, 3923, 3929, 3931, 3943, 3947, 3967, 3989, 
4001, 4003, 4007, 4013, 4019, 4021, 4027, 4049, 4051, 4057, 
4073, 4079, 4091, 4093, 4099, 4111, 4127, 4129, 4133, 4139, 
4153, 4157, 4159, 4177, 4201, 4211, 4217, 4219, 4229, 4231, 
4241, 4243, 4253, 4259, 4261, 4271, 4273, 4283, 4289, 4297, 
4327, 4337, 4339, 4349, 4357, 4363, 4373, 4391, 4397, 4409, 
4421, 4423, 4441, 4447, 4451, 4457, 4463, 4481, 4483, 4493, 
4507, 4513, 4517, 4519, 4523, 4547, 4549, 4561, 4567, 4583, 
4591, 4597, 4603, 4621, 4637, 4639, 4643, 4649, 4651, 4657, 
4663, 4673, 4679, 4691, 4703, 4721, 4723, 4729, 4733, 4751, 
4759, 4783, 4787, 4789, 4793, 4799, 4801, 4813, 4817, 4831, 
4861, 4871, 4877, 4889, 4903, 4909, 4919, 4931, 4933, 4937, 
4943, 4951, 4957, 4967, 4969, 4973, 4987, 4993, 4999, 5003, 
5009, 5011, 5021, 5023, 5039, 5051, 5059, 5077, 5081, 5087, 
5099, 5101, 5107, 5113, 5119, 5147, 5153, 5167, 5171, 5179, 
5189, 5197, 5209, 5227, 5231, 5233, 5237, 5261, 5273, 5279, 
5281, 5297, 5303, 5309, 5323, 5333, 5347, 5351, 5381, 5387, 
5393, 5399, 5407, 5413, 5417, 5419, 5431, 5437, 5441, 5443, 
5449, 5471, 5477, 5479, 5483, 5501, 5503, 5507, 5519, 5521, 
5527, 5531, 5557, 5563, 5569, 5573, 5581, 5591, 5623, 5639, 
5641, 5647, 5651, 5653, 5657, 5659, 5669, 5683, 5689, 5693, 
5701, 5711, 5717, 5737, 5741, 5743, 5749, 5779, 5783, 5791, 
5801, 5807, 5813, 5821, 5827, 5839, 5843, 5849, 5851, 5857, 
5861, 5867, 5869, 5879, 5881, 5897, 5903, 5923, 5927, 5939, 
5953, 5981, 5987, 6007, 6011, 6029, 6037, 6043, 6047, 6053, 
6067, 6073, 6079, 6089, 6091, 6101, 6113, 6121, 6131, 6133, 
6143, 6151, 6163, 6173, 6197, 6199, 6203, 6211, 6217, 6221, 
6229, 6247, 6257, 6263, 6269, 6271, 6277, 6287, 6299, 6301, 
6311, 6317, 6323, 6329, 6337, 6343, 6353, 6359, 6361, 6367, 
6373, 6379, 6389, 6397, 6421, 6427, 6449, 6451, 6469, 6473, 
6481, 6491, 6521, 6529, 6547, 6551, 6553, 6563, 6569, 6571, 
6577, 6581, 6599, 6607, 6619, 6637, 6653, 6659, 6661, 6673, 
6679, 6689, 6691, 6701, 6703, 6709, 6719, 6733, 6737, 6761, 
6763, 6779, 6781, 6791, 6793, 6803, 6823, 6827, 6829, 6833, 
6841, 6857, 6863, 6869, 6871, 6883, 6899, 6907, 6911, 6917, 
6947, 6949, 6959, 6961, 6967, 6971, 6977, 6983, 6991, 6997, 
7001, 7013, 7019, 7027, 7039, 7043, 7057, 7069, 7079, 7103, 
7109, 7121, 7127, 7129, 7151, 7159, 7177, 7187, 7193, 7207,
7211, 7213, 7219, 7229, 7237, 7243, 7247, 7253, 7283, 7297, 
7307, 7309, 7321, 7331, 7333, 7349, 7351, 7369, 7393, 7411, 
7417, 7433, 7451, 7457, 7459, 7477, 7481, 7487, 7489, 7499, 
7507, 7517, 7523, 7529, 7537, 7541, 7547, 7549, 7559, 7561, 
7573, 7577, 7583, 7589, 7591, 7603, 7607, 7621, 7639, 7643, 
7649, 7669, 7673, 7681, 7687, 7691, 7699, 7703, 7717, 7723, 
7727, 7741, 7753, 7757, 7759, 7789, 7793, 7817, 7823, 7829, 
7841, 7853, 7867, 7873, 7877, 7879, 7883, 7901, 7907, 7919};
		
	for( row = 1; row <= numberOfRows; row++ ){
		rowLength = numberOfNz( bitPtr[row] );
		matRow = matPtr[row];

		max = 0;
		for( i = 1; i <= rowLength; i++ ){
			if( labs(matRow[i]) > (long) max ){
			       max = (UINT32) labs(matRow[i]);
			}
	 	}		
		i = 0;
		while(1){
			nextPrime = primes[i];
			j = 1;
			while( matRow[j] % (XTYPE) nextPrime == 0 ){
				j++;
				if( j > rowLength ){
					break;
				}
			}
			if( j == (rowLength + 1) ){
				for( k = 1; k <= rowLength; k++ ){
					matRow[k] /= (XTYPE) nextPrime;
				}
			}
			else{
				i++;
			}
			
			if( (nextPrime > max/2) || (i >= 1000) ){
				break;
			}
		}/*while(1)*/

	}/*for(row)*/


} /*reduceRow()*/

/***************************************************************************************************************************************
 *
 * 	Function: 	findColumn()
 *
 * 	Description:	This function returns the column from the circular queue which has the smallest value of
 * 			pos * neg;
 * 			
 *
 *	Parameters:	b1Ptr -- the bit array
 * 	     		m1Ptr -- the full matrix
 * 			qSize -- size of the queue
 *
 *	Date:		December 14, 2004
 *
 * 	Author:		Steve Bell
 *
 * ***************************************************************************************************************************************/


UINT32 findColumn(BIT_TYPE** b1Ptr, XTYPE** m1Ptr, UINT32 qSize)
{
	UINT32 minCol, dummy = 0, minNR, colNumber; 
	UINT32 bits, row, col, xtype; /*dummy variables*/
	BOOLEAN STOP = my_false;
	/*these are used to find a particular column in the bit array*/ 
	UINT32 colSlot;
	BIT_TYPE colBit, xWord, testBit;
	UINT32 nzCol, diff; /*these are used to find non zero rows */
	
    /*for a particular column, these keep track of which rows have zero, nonZero, positive, negative
	 * entries*/	
	UINT32 zero, neg, pos, nonZero, cnewRows, countNz, countCol; 
	minNR = MAXBITS; 
	dummy = 0; /*used to find minimum column*/
	
	/*find column which has the minimum number of positive rows * number of negative rows*/
	for( colNumber = 1; colNumber <= qSize; colNumber++ ){
		GetWord( (UINT32*) &col); 
		/*find column in bit array*/
		colSlot = (col - 1)/BIT_TYPESIZE  + 1;
		if( (col - 1) % BIT_TYPESIZE  != 0 ){
			colBit =  HIGH_BIT >> ( (col - 1) % BIT_TYPESIZE );
		}
		else{
			colBit = HIGH_BIT;
		}
		neg = 0;
		pos = 0;
		zero = 0;
		nonZero = 0;
				
		for( row = 1; row <= T1nrows; row++ ){
			if( b1Ptr[row] == NULL ){
				continue;
			}
			if( (b1Ptr[row][colSlot] & colBit) != 0 ){
			
				/*find column in sparse row vector, corresponding to col*******************************************/
				STOP = my_false;
				countNz = 0;
				countCol = 1;

				for( xtype = 1; ((xtype <= NUMBER_OF_XTYPES) && (STOP == my_false)); xtype++ ){
					testBit = HIGH_BIT;
					xWord = b1Ptr[row][xtype];
					for( bits = 1; bits <= BIT_TYPESIZE; bits++ ){			
						if( (testBit & xWord) != 0 ){
							countNz++;
						}
						testBit = testBit >> 1;
						countCol++;
						if( countCol > col ){
							STOP = my_true;
							break;
						}
					}
				}
				if( STOP == my_false ){
					testBit = HIGH_BIT;
					xWord = b1Ptr[row][xtype];
					for( bits = 1;bits <= NUMBER_OF_RESIDS; bits++ ){		
						if( (testBit & xWord) != 0 ){
							countNz++;
						}
						testBit = testBit >> 1;
						countCol++;
						if( countCol > col ){
							break;
						}
					}
				}

				/*end find column*******************************************************************************/

				nzCol = countNz;
				/*find rows with negative entries*/
				if( m1Ptr[row][nzCol] < 0 ){
					neg++;
				}
				/*find rows with positive entries*/
				else{ /*else > 0*/
					pos++;
				}
			} /*if( (b1Ptr[row][colSlot] & colBit) != 0 ) */
			else{
				zero++;
			}
		} /*for(row)*/
		cnewRows = neg * pos;
		diff = labs(neg - pos);
		if( diff == 0 ){
			diff = 1;
		}
		if( cnewRows > 0 ){
			cnewRows = pos * neg;
		}
		if( (cnewRows < minNR) || (T1nrows == 0) ){
			minNR = cnewRows;
			if( dummy == 1 ){
				PutWord(minCol);
				minCol = col;/*don't put the minimum column back into the queue*/
			}
			else{/*first time*/
				dummy = 1;
				minCol = col;
			}
		}
		else{
			PutWord(col);
		}
	} /*for(colNumber)*/

	col = minCol;

	return(col);
}/*findColumn()*/

